import unittest
from unittest.mock import Mock, patch, PropertyMock
import ctypes
import numpy as np
import sys

# Existing import of ps4000a might cause a disturbance
if "picosdk.ps4000a" in sys.modules:
    del sys.modules["picosdk.ps4000a"]

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_TimeoutException, QMI_InstrumentException, QMI_UsageException

# For not causing an error by missing picosdk library in the pipeline tests, we mock also the picosdk existence
import tests.instruments.picotech.picosdk_stub
from tests.instruments.picotech.ps3000a_stub import ps3000a
sys.modules['picosdk'] = tests.instruments.picotech.picosdk_stub
sys.modules['picosdk.ps3000a'] = tests.instruments.picotech.ps3000a_stub
from qmi.instruments.picotech import picoscope3403, ChannelCoupling, TriggerEdge
from qmi.instruments.picotech import _picoscope

patcher = patch("picosdk.ps3000a", autospec=True)
picoscope3403._ps = patcher.start()


class PicoscopeMethodsTestCase(unittest.TestCase):

    def setUp(self) -> None:
        _picoscope._ps = ps3000a
        self._picoscope = picoscope3403.PicoTech_PicoScope3403(
            QMI_Context('test_picoscope_3403'), 'PicoScope3403', "GR956/0069"
        )
        # Mock the PicoScope "open" function to be able to open the virtual unit.
        self._picoscope._ps = PropertyMock(autospec=True)
        self._picoscope.open()

    def tearDown(self) -> None:
        self._picoscope.close()
        picoscope3403._ps.reset_mock()

    def test_get_serial_number(self):
        """Test query serial number. Mock can only return error code, need to patch the input string."""
        expected = "s1234"
        self._picoscope._ps_attr.ps3000aGetUnitInfo = Mock(return_value=0)
        with patch('ctypes.create_string_buffer', side_effect=[ctypes.c_char_p(bytes(expected.encode()))]):
            sn = self._picoscope.get_serial_number()

        self.assertEqual(sn, expected)

    def test_device_variant(self):
        """Test query device variant. Mock can only return error code, need to patch the input string."""
        expected = "ps3403"
        _picoscope._ps.PICO_INFO["PICO_VARIANT_INFO"] = "3403"
        self._picoscope._ps_attr.ps3000aGetUnitInfo = Mock(return_value=0)
        with patch('ctypes.create_string_buffer', side_effect=[ctypes.c_char_p(bytes(expected.encode()))]):
            variant = self._picoscope.get_device_variant()

        self.assertEqual(variant, expected)

    def test_input_ranges(self):
        """Test query input ranges."""
        sel = range(11)
        vol = [i / 10.0 for i in range(11)]
        expected = dict(zip(sel, vol))

        _picoscope._ps.PICO_VOLTAGE_RANGE = expected
        ranges = self._picoscope.get_input_ranges()
        self.assertDictEqual(expected, ranges)

    def test_set_channel_ok(self):
        """Test setting channel values with no invalid input values."""
        self._picoscope.set_channel(1, True, ChannelCoupling.AC, 0, 1.0)

    def test_set_channel_nok_invalid_channel_index(self):
        """Test setting channel values with invalid channel index numbers."""
        # channel index < 0
        with self.assertRaises(ValueError):
            self._picoscope.set_channel(-1, True, ChannelCoupling.AC, 0, 1.0)

        # channel index >= NUM_CHANNELS
        with self.assertRaises(ValueError):
            self._picoscope.set_channel(picoscope3403.PicoTech_PicoScope3403.NUM_CHANNELS, True,
                                        ChannelCoupling.AC, 0, 1.0)

    def test_set_channel_nok_invalid_input_range(self):
        """Test setting channel values with invalid input range numbers."""
        # input_range < 0
        with self.assertRaises(ValueError):
            self._picoscope.set_channel(0, True, ChannelCoupling.AC, -1, 1.0)

        # input_range >= NUM_INPUT_RANGES
        with self.assertRaises(ValueError):
            self._picoscope.set_channel(1, True, ChannelCoupling.AC,
                                        picoscope3403.PicoTech_PicoScope3403.NUM_INPUT_RANGES, 1.0)

    def test_set_trigger_ok(self):
        """Test setting trigger with valid input values."""
        self._picoscope.set_trigger(False, 1, -10000, TriggerEdge.RISING_OR_FALLING)

    def test_set_trigger_nok_invalid_channel_index(self):
        """Test setting trigger with invalid channel indexes."""
        # channel index < 0
        with self.assertRaises(ValueError):
            self._picoscope.set_trigger(True, -1, -10000, TriggerEdge.RISING)

        # channel index >= NUM_CHANNELS
        with self.assertRaises(ValueError):
            self._picoscope.set_trigger(True, picoscope3403.PicoTech_PicoScope3403.NUM_CHANNELS, -10000,
                                        TriggerEdge.FALLING)

    def test_set_trigger_nok_invalid_threshold(self):
        """Test setting trigger with invalid threshold values."""
        # channel index < MIN_SAMPLE_VALUE
        with self.assertRaises(ValueError):
            self._picoscope.set_trigger(True, 1, picoscope3403.PicoTech_PicoScope3403.MIN_SAMPLE_VALUE - 1,
                                        TriggerEdge.RISING)

        # channel index > MAX_SAMPLE_VALUE
        with self.assertRaises(ValueError):
            self._picoscope.set_trigger(False, 2, picoscope3403.PicoTech_PicoScope3403.MAX_SAMPLE_VALUE + 1,
                                        TriggerEdge.FALLING)

    def test_set_disable_trigger_ok(self):
        """Test disabling all triggers."""
        self._picoscope.disable_trigger()

    def test_stop(self):
        """Test stop function."""
        self._picoscope.stop()

    def test_run_block(self):
        """Test the run_block function."""
        self._picoscope._ps_attr.ps3000aGetTimebase2 = Mock(return_value=0)
        self._picoscope._ps_attr.ps3000aRunBlock = Mock(return_value=0)
        self._picoscope.run_block(10, 10, 100)
        picoscope3403._ps.ps3000aGetTimebase2.assert_called_once()
        picoscope3403._ps.ps3000aRunBlock.assert_called_once()

    def test_is_block_ready(self):
        """Test the if_block_ready function. Note that we cannot check for the case is_block_ready returns True."""
        picoscope3403._ps.ps3000aGetTimebase2 = Mock(return_value=0)
        picoscope3403._ps.ps3000aRunBlock = Mock(return_value=0)
        # We need to run first the run_block to be able to run the is_block_ready
        self._picoscope.run_block(10, 10, 100)
        self._picoscope.is_block_ready()

        picoscope3403._ps.ps3000aGetTimebase2.assert_called_once()
        picoscope3403._ps.ps3000aRunBlock.assert_called_once()

    def test_wait_block_ready(self):
        """Test that the wait_block_ready returns normally without timeout."""
        self._picoscope._ps_attr.ps3000aGetTimebase2 = Mock(return_value=0)
        self._picoscope._ps_attr.ps3000aRunBlock = Mock(return_value=0)
        # We need to run first the run_block to be able to run the is_block_ready
        self._picoscope.run_block(10, 10, 100)
        with patch('ctypes.c_int16', side_effect=[ctypes.c_int16(1)]):
            self._picoscope.wait_block_ready(1.0)

        picoscope3403._ps.ps3000aGetTimebase2.assert_called_once()
        picoscope3403._ps.ps3000aRunBlock.assert_called_once()

    def test_wait_block_ready_timeout(self):
        """Test the wait_block_ready function with timeout. Note that we can check only that timeout occurs."""
        self._picoscope._ps_attr.ps3000aGetTimebase2 = Mock(return_value=0)
        self._picoscope._ps_attr.ps3000aRunBlock = Mock(return_value=0)
        # We need to run first the run_block to be able to run the is_block_ready
        self._picoscope.run_block(10, 10, 100)
        with self.assertRaises(QMI_TimeoutException):
            self._picoscope.wait_block_ready(0.5)

        picoscope3403._ps.ps3000aGetTimebase2.assert_called_once()
        picoscope3403._ps.ps3000aRunBlock.assert_called_once()

    def test_wait_block_ready_invalid_timeout(self):
        """Test the wait_block_ready function returns an error at invalid timeout."""
        with self.assertRaises(ValueError):
            self._picoscope.wait_block_ready(-1.0)

    def test_get_block_data_ok(self):
        """Test the get_block_data function returns data for channels 1 & 3. Only 0 data."""
        channels = [1, 3]
        samples_before = 5
        samples_after = 10
        self._picoscope._ps_attr.ps3000aGetTimebase2 = Mock(return_value=0)
        self._picoscope._ps_attr.ps3000aRunBlock = Mock(return_value=0)
        self._picoscope._ps_attr.ps3000aSetDataBuffer = Mock(return_value=0)
        self._picoscope._ps_attr.ps3000aGetValues = Mock(return_value=0)
        # We need to run first the run_block to be able to run the get_block_data correctly
        self._picoscope.run_block(samples_before, samples_after, 100)
        samples, timebase_interval_ns, overrange = self._picoscope.get_block_data(channels)
        # Assert that we get back samples and overrange for two channels and that the number of channels match the
        # number of channels given in the run_block command.
        self.assertEqual(len(channels), len(samples))
        self.assertEqual(len(channels), len(overrange))
        self.assertEqual(samples_before + samples_after, len(samples[0]))

    def test_get_block_data_nok_invalid_channel_indexes(self):
        """Test the get_block_data function raises an error with wrong number of samples."""
        channels = [-1, 3]
        # channel index < 0
        with self.assertRaises(ValueError):
            self._picoscope.get_block_data(channels)

        channels = [1, picoscope3403.PicoTech_PicoScope3403.NUM_CHANNELS]
        # channel index >= NUM_CHANNELS
        with self.assertRaises(ValueError):
            self._picoscope.get_block_data(channels)

    def test_get_block_data_nok_wrong_number_of_samples(self):
        """Test the get_block_data function raises an error with wrong number of samples."""
        channels = [1, 3]
        samples_before = 10
        samples_after = 10
        picoscope3403._ps.ps3000aGetTimebase2 = Mock(return_value=0)
        picoscope3403._ps.ps3000aRunBlock = Mock(return_value=0)
        self._picoscope._ps_attr.ps3000aSetDataBuffer = Mock(return_value=0)
        self._picoscope._ps_attr.ps3000aGetValues = Mock(return_value=0)
        # We need to run first the run_block to be able to run the get_block_data correctly
        self._picoscope.run_block(samples_before, samples_after, 100)
        with patch('ctypes.c_uint32', side_effect=[ctypes.c_uint32(15)]):
            with self.assertRaises(QMI_InstrumentException):
                self._picoscope.get_block_data(channels)
    
    def test_acquire_by_trigger_and_get_data_ok_interval_1_2_4ns(self):
        """Test that the acquire_by_trigger_and_get_data returns data with correct settings
        for sampling intervals of 1, 2 and 4ns."""
        # prepare
        channels = [[2], [0, 2], [0, 2]]
        max_val = [0.05, 0.2, 0.1]  # List needs to minimally be length of the largest channel number + 1 in channels
        couplings = [
            ChannelCoupling.AC,
            ChannelCoupling.DC,
            ChannelCoupling.DC,
            ]  # List needs to minimally be length of the largest channel number + 1 in channels
        trigger_channel = 2
        trigger_level = 0.02
        sampling_intervals = [1, 2, 4]  # ns
        samples = 100
        self._picoscope._ps_attr.ps3000aSetChannel = Mock(return_value=0)
        self._picoscope._ps_attr.ps3000aSetSimpleTrigger = Mock(return_value=0)
        self._picoscope._ps_attr.ps3000aStop = Mock(return_value=0)
        picoscope3403._ps.ps3000aGetTimebase2 = Mock(return_value=0)
        picoscope3403._ps.ps3000aRunBlock = Mock(return_value=0)
        self._picoscope._ps_attr.ps3000aIsReady = Mock(return_value=0)
        self._picoscope._ps_attr.ps3000aSetDataBuffer = Mock(return_value=0)
        self._picoscope._ps_attr.ps3000aGetValues = Mock(return_value=0)

        for i in range(3):
            time_span = sampling_intervals[i] * samples  # ns
            with patch('ctypes.c_int16', side_effect=[ctypes.c_int16(1), ctypes.c_int16(0)]):
                with patch('numpy.empty', side_effect=[np.empty((len(channels), 0), dtype=np.int16)]):

                    tt, voltgs = self._picoscope.acquire_by_trigger_and_get_data(
                        channels[i],
                        max_val,
                        couplings,
                        trigger_channel,
                        trigger_level,
                        sampling_intervals[i],
                        time_span,
                        )
                    # Assert that time-trace length is the same as number of samples
                    self.assertEqual(samples, len(tt))
                    # Did not manage to patch the voltages output, so cannot assert that.

    def test_acquire_by_trigger_and_get_data_ok_interval_8ns(self):
        """Test that the acquire_by_trigger_and_get_data returns data without timeout."""
        # prepare
        channels = [0, 2]
        max_val = [0.5, 2.0, 1.0]  # List needs to minimally be length of the largest channel number + 1 in channels
        couplings = [
            ChannelCoupling.AC,
            ChannelCoupling.DC,
            ChannelCoupling.DC,
            ]  # List needs to minimally be length of the largest channel number + 1 in channels
        trigger_channel = 2
        trigger_level = 0.2
        sampling_interval = 8  # ns
        samples = 100
        time_span = sampling_interval * samples  # ns
        self._picoscope._ps_attr.ps3000aSetChannel = Mock(return_value=0)
        self._picoscope._ps_attr.ps3000aSetSimpleTrigger = Mock(return_value=0)
        self._picoscope._ps_attr.ps3000aStop = Mock(return_value=0)
        picoscope3403._ps.ps3000aGetTimebase2 = Mock(return_value=0)
        picoscope3403._ps.ps3000aRunBlock = Mock(return_value=0)
        self._picoscope._ps_attr.ps3000aIsReady = Mock(return_value=0)
        self._picoscope._ps_attr.ps3000aSetDataBuffer = Mock(return_value=0)
        self._picoscope._ps_attr.ps3000aGetValues = Mock(return_value=0)

        with patch('ctypes.c_int16', side_effect=[ctypes.c_int16(1), ctypes.c_int16(0)]):
            with patch('numpy.empty', side_effect=[np.empty((len(channels), 0), dtype=np.int16)]):

                tt, voltgs = self._picoscope.acquire_by_trigger_and_get_data(
                    channels,
                    max_val,
                    couplings,
                    trigger_channel,
                    trigger_level,
                    sampling_interval,
                    time_span,
                    )
                # Assert that time-trace length is the same as number of samples
                self.assertEqual(samples, len(tt))
                # Did not manage to patch the voltages output, so cannot assert that.

    def test_acquire_no_trigger_and_get_data_ok(self):
        """Test that the acquire_by_trigger_and_get_data returns data without trigger."""
        # prepare
        channels = [0, 2]
        max_val = [5.0, 20.0, 10.0]  # List needs to minimally be length of the largest channel number + 1 in channels
        couplings = [
            ChannelCoupling.AC,
            ChannelCoupling.DC,
            ChannelCoupling.DC,
            ]  # List needs to minimally be length of the largest channel number + 1 in channels
        trigger_channel = None
        trigger_level = 2.0
        sampling_interval = 8  # ns
        samples = 100
        time_span = sampling_interval * samples  # ns
        self._picoscope._ps_attr.ps3000aSetChannel = Mock(return_value=0)
        self._picoscope._ps_attr.ps3000aSetSimpleTrigger = Mock(return_value=0)
        self._picoscope._ps_attr.ps3000aStop = Mock(return_value=0)
        picoscope3403._ps.ps3000aGetTimebase2 = Mock(return_value=0)
        picoscope3403._ps.ps3000aRunBlock = Mock(return_value=0)
        self._picoscope._ps_attr.ps3000aIsReady = Mock(return_value=0)
        self._picoscope._ps_attr.ps3000aSetDataBuffer = Mock(return_value=0)
        self._picoscope._ps_attr.ps3000aGetValues = Mock(return_value=0)

        with patch('ctypes.c_int16', side_effect=[ctypes.c_int16(1), ctypes.c_int16(0)]):
            with patch('numpy.empty', side_effect=[np.empty((len(channels), 0), dtype=np.int16)]):

                tt, voltgs = self._picoscope.acquire_by_trigger_and_get_data(
                    channels,
                    max_val,
                    couplings,
                    trigger_channel,
                    trigger_level,
                    sampling_interval,
                    time_span,
                    )
                # Assert that time-trace length is the same as number of samples
                self.assertEqual(samples, len(tt))
                # Did not manage to patch the voltages output, so cannot assert that.

    def test_acquire_by_trigger_and_get_data_nok_with_timeout(self):
        """Test that the acquire_by_trigger_and_get_data excepts with timeout."""
        # prepare
        channels = [0, 2]
        max_val = [5.0, 2.0, 1.0]  # List needs to minimally be length of the largest channel number + 1 in channels
        couplings = [
            ChannelCoupling.AC,
            ChannelCoupling.DC,
            ChannelCoupling.DC,
            ]  # List needs to minimally be length of the largest channel number + 1 in channels
        trigger_channel = 2
        trigger_level = 0.2
        sampling_interval = 4  # ns
        time_span = sampling_interval * 100  # ns

        picoscope3403._ps.ps3000aGetTimebase2 = Mock(return_value=0)
        picoscope3403._ps.ps3000aRunBlock = Mock(return_value=0)
        self._picoscope._ps_attr.ps3000aIsReady = Mock(return_value=0)
        with self.assertRaises(QMI_TimeoutException):
            self._picoscope.acquire_by_trigger_and_get_data(
                channels,
                max_val,
                couplings,
                trigger_channel,
                trigger_level,
                sampling_interval,
                time_span,
                )

    def test_acquire_by_trigger_and_get_data_nok_invalid_channel_index(self):
        """Test trying to acquire data with invalid channel indexes."""
        channels = [-1, picoscope3403.PicoTech_PicoScope3403.NUM_CHANNELS]
        # channel index < 0, channel index >= NUM_CHANNELS
        for channel in channels:
            with self.assertRaises(ValueError):
                self._picoscope.acquire_by_trigger_and_get_data(
                    [channel],
                    [1.0],
                    [ChannelCoupling.AC],
                    0,
                    0.2,
                    250,
                    125,
                    )

    def test_acquire_by_trigger_and_get_data_nok_invalid_trigger_index(self):
        """Test trying to acquire data with invalid trigger channel indexes."""
        trigger_channels = [-1, picoscope3403.PicoTech_PicoScope3403.NUM_CHANNELS]
        # trigger channel index < 0, trigger channel index >= NUM_CHANNELS
        for trigger_channel in trigger_channels:
            with self.assertRaises(ValueError):
                self._picoscope.acquire_by_trigger_and_get_data(
                    [0],
                    [1.0],
                    [ChannelCoupling.AC],
                    trigger_channel,
                    0.2,
                    250,
                    125,
                    )

        trigger_channel = 0
        # trigger channel not in the channels list
        with self.assertRaises(ValueError):
            self._picoscope.acquire_by_trigger_and_get_data(
                [c for c in range(1, picoscope3403.PicoTech_PicoScope3403.NUM_CHANNELS)],
                [1.0],
                [ChannelCoupling.AC],
                trigger_channel,
                0.2,
                250,
                125,
                )

    def test_acquire_by_trigger_and_get_data_nok_incorrect_channels_to_timebase(self):
        """Test trying to acquire data with invalid number of channels to time-bases 0 and 1."""
        channels = [[0, 1], [1, 2, 3]]
        sampling_intervals = [1, 2]  # For time-bases 0 and 1
        for i in range(2):
            with self.assertRaises(QMI_UsageException):
                self._picoscope.acquire_by_trigger_and_get_data(
                    channels[i],
                    [1.0, 1.0],
                    [ChannelCoupling.AC],
                    1,
                    0.2,
                    sampling_intervals[i],
                    80,
                    )

    def test_acquire_by_trigger_and_get_data_nok_incorrect_sampling_intervals(self):
        """Test trying to acquire data with invalid sampling_interval values."""
        sampling_intervals = [0, 3, 7, 15, 19, 13235]
        for sampling_interval in sampling_intervals:
            with self.assertRaises(QMI_UsageException):
                self._picoscope.acquire_by_trigger_and_get_data(
                    [0, 1, 2, 3],
                    [1.0],
                    [ChannelCoupling.AC],
                    1,
                    0.2,
                    sampling_interval,
                    125,
                    )

    def test_get_sampling_interval_ok(self):
        """Test that the get_sampling_interval returns correct values for various inputs in range 0...2**32 - 1."""
        input_values = list(range(6)) + [(96 / 8) + 2, 2**32 - 1]
        expected = [1, 2, 4] + list(range(8, 4*8, 8)) + [96, (2**32 - 3) / 125E-3]
        for i in range(len(input_values)):
            result = self._picoscope.get_sampling_interval(input_values[i])
            self.assertEqual(expected[i], int(result))

    def test_get_sampling_interval_out_of_range(self):
        """Test that the get_sampling_interval returns 0 for inputs < 0 and >= 2**32."""
        input_values = [-1, 2**32]
        expected = [0, 0]
        for i in range(len(input_values)):
            result = self._picoscope.get_sampling_interval(input_values[i])
            self.assertEqual(expected[i], result)


if __name__ == "__main__":
    unittest.main()



"""Extension of the QMI_UsbTmcTransport class utilizing usb.core with libusb1 backend and pyvisa packages."""

from typing import Optional

import os, sys
from qmi.core.exceptions import QMI_TimeoutException, QMI_TransportDescriptorException, QMI_InstrumentException,\
    QMI_RuntimeException
from qmi.core.transport import QMI_UsbTmcTransport

if sys.platform == "win32":
    import usb
    import usb.core
    from usb.backend import libusb1
    try:
        backend = libusb1.get_backend(find_library=lambda x: os.getenv("LIBUSBPATH"))
        usb.core.find(backend=backend)

    except usb.core.NoBackendError as exc:
        raise QMI_RuntimeException("LIBUSBPATH environment variable must be set for libusb") from exc

import pyvisa
import pyvisa.errors


class QMI_VisaUsbTmcTransport(QMI_UsbTmcTransport):

    def __init__(self, vendorid: int, productid: int, serialnr: str) -> None:
        super().__init__(vendorid, productid, serialnr)
        self._device: Optional[pyvisa.ResourceManager] = None

    def _open_transport(self) -> None:
        visa_resource = f"USB::0x{self.vendorid:04x}::0x{self.productid:04x}::{self.serialnr}::INSTR"
        rm = pyvisa.ResourceManager()
        try:
            self._device = rm.open_resource(visa_resource)

        except ValueError as exc:
            if "install a suitable backend" in str(exc):
                raise QMI_TransportDescriptorException("LIBUSBPATH environment variable must be set for libusb") \
                    from exc

            else:
                raise QMI_InstrumentException(exc)

    @property
    def _safe_device(self) -> pyvisa.ResourceManager:
        """ The _safe_device property should be used inside the QMI_Transport code if-and-only-if we are 100% sure that
        the _device attribute is not None.

        This aids in static typechecking, since whereas the type of _device is Optional[T], the result of this method is
        guaranteed to be of type T. It is a QMI-internal bug if this property is used in case _device is None. In that
        case, we raise an AssertionError, and we hope the users will complain to us so we can fix the bug in the
        library.

        Raises: AssertionError: in case the property is used when the underlying value of _device is None.

        Returns: The value of _device, if it is not None. """
        assert self._device is not None
        return self._device

    def write(self, data: bytes) -> None:
        self._check_is_open()

        self._safe_device.timeout = int(1000 * self.WRITE_TIMEOUT)
        self._safe_device.write_raw(data)

    def _read_message(self, timeout):
        self._safe_device.timeout = 1 + int(timeout * 1000)
        try:
            data = self._safe_device.read_raw()
        except pyvisa.errors.VisaIOError as exc:
            if exc.error_code == pyvisa.errors.VI_ERROR_TMO:
                raise QMI_TimeoutException(str(exc)) from exc
            else:
                raise
        return data

    @staticmethod
    def list_resources():
        rm = pyvisa.ResourceManager()
        resources = rm.list_resources()
        formatted_resources = QMI_UsbTmcTransport._format_resources(resources)
        return formatted_resources



"""Publish/subscribe messaging mechanism for QMI.

The publish/subscribe mechanism is suitable for situations where data
is spontaneously produced (i.e. not in response to a request). Such data
may be sent to other QMI objects by `publishing a signal`.

Tasks which want to act on signals may create a signal receiver and
`subscribe` it to the corresponding publisher.

Declaring signals
#################

Any subclass of ``QMI_RpcObject`` (such as instruments and tasks) can publish signals.
In order to publish a signal, the signal must be declared in the RPC object class definition.
This is done by creating a `class attribute` and initializing it with an instance of ``QMI_Signal``.

For example::

    class MyTask(QMI_Task):

        # Declare a signal with name "sig_alice" which takes two parameters, both integers.
        sig_alice = QMI_Signal([int, int])

        # Declare a signal with name "sig_bob" which takes one string parameter.
        sig_bob = QMI_Signal([str])

        # ... rest of class definition

Publishing a signal
###################

An RPC object can publish any of the signals it has declared.
Publishing a signal means broadcasting a message through the QMI network.
This message is automatically received by all signal receivers that are subscribed to the signal.

For example, the task `MyTask` could publish its signals as follows::

    class MyTask(QMI_Task):
        sig_alice = QMI_Signal([int, int])
        sig_bob = QMI_Signal([str])

        def run(self):

            # Publish "sig_alice"
            self.sig_alice.publish(11, 105)

            # Publish "sig_bob"
            self.sig_bob.publish("hello receiver")

            # Publish "sig_alice" again with different parameters.
            self.sig_alice.publish(5, 6)

Subscribing to signals
######################

Any Python script, task or function can subscribe to signals.
Subscribing to a signal requires a `proxy` for the object that publishes the signal,
and a `signal receiver`.

For example, the following code creates an instance of `MyTask`.
It then uses the proxy to subscribe a receiver to the signal `sig_alice`::

    # Create task (and get a proxy to the new task).
    my_task_proxy = qmi.make_task("my_task", MyTask)

    # Create signal receiver.
    receiver = QMI_SignalReceiver()

    # Subscribe receiver to the signal "sig_alice".
    my_task_proxy.sig_alice.subscribe(receiver)

You can always subscribe to signals that are published in the same QMI context
(i.e. the same Python program).
It is also possible to subscribe to signals that are published by
another Python program, provided that this program is a peer context in
the QMI network, and a connection to this peer context has been established.

A receiver can be subscribed to multiple signals, and multiple
receivers can be subscribed to the same signal.

Receiving signals
#################

A signal receiver keeps received signals in an internal queue.
Each time a signal gets published to which the receiver is currently subscribed,
it adds a `ReceivedSignal` record to its queue.

In order to process a received signal, you can call `get_next_signal()` to
get a `ReceivedSignal` record from the receive queue.
If there are received signals in the queue, this function returns the oldest
received signal.
Otherwise, if the queue is empty, the function will either wait until
the next signal is received, or raise a `QMI_TimeoutException`.

For example::

    # Let's assume that "receiver" is already subscribed to "sig_alice"
    # and the signal has been published via "sig_alice.publish(11, 105)".
    try:
        sig = receiver.get_next_signal(timeout=1.0)
        print("Received signal", sig.signal_name, "with arguments", sig.args[0], sig.args[1])
          # will print something like "Received signal sig_alice with arguments 11 105"
    except QMI_TimeoutException:
        print("No signal was received within 1 second")

Reference
#########
"""

import logging
import threading
from collections import deque

import typing
from typing import Any, Callable, Dict, List, NamedTuple, Optional, Set, Tuple, Type

from qmi.core.exceptions import (
    QMI_TimeoutException, QMI_MessageDeliveryException,
    QMI_RuntimeException, QMI_SignalSubscriptionException, QMI_UsageException)
from qmi.core.messaging import (
    QMI_Message, QMI_RequestMessage, QMI_ReplyMessage, QMI_ErrorReplyMessage,
    QMI_MessageHandler, QMI_MessageHandlerAddress)
from qmi.core.util import is_valid_object_name


# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class SignalDescription(NamedTuple):
    """Description of signal type, specifying its name and arguments.

    Attributes:
        name: Signal name.
        arg_types: Tuple of Python types (e.g. `int` or `str`), describing
            the parameters that will be passed in each published instance of
            this signal type.
    """
    name: str
    arg_types: Tuple[Type, ...]


class ReceivedSignal(NamedTuple):
    """Signal instance received by a `QMI_SignalReceiver`.

    When a `QMI_SignalReceiver` receives a signal, it stores an instance
    of `ReceivedSignal` in its receive queue. The application can retrieve
    this instance and handle it.

    Attributes:
        publisher_context: Name of the context that published the signal.
        publisher_name: Name of the object that published the signal.
        signal_name: Name of the signal.
        args: List of signal arguments.
        receiver_seqnr: Sequence number assigned by the `QMI_SignalReceiver`.
            This sequence number can be used to detect dropped signals in
            case of overflow of the receive queue.
    """
    publisher_context:  str
    publisher_name: str
    signal_name: str
    args: tuple
    receiver_seqnr: int


class QMI_Signal:
    """Marker for signal declarations in an RpcObject or Task.

    An RpcObject or Task that wants to publish a signal, creates
    a class attribute with the name of the signal and initializes
    it with an instance of `QMI_Signal`.

    For example::

        class MyObject(QMI_RpcObject):
            my_signal = QMI_Signal([int, str])
    """
    __slots__ = ("arg_types",)

    def __init__(self, arg_types: List[Type]) -> None:
        """Declare a new signal type.

        Parameters:
            arg_types: List of Python types, describing the parameters that
                will be passed in each published instance of this signal type.
        """
        self.arg_types = tuple(arg_types)

    def publish(self, *args: Any) -> None:
        """Publish this signal.

        This method will be implemented by `QMI_RegisteredSignal`.
        It is declared here for static type checking.
        """
        raise QMI_RuntimeException("QMI_Signal has not been registered and converted to QMI_RegisteredSignal.")


class QMI_RegisteredSignal:
    """Represents a named signal, bound to a specific `QMI_RpcObject` instance.

    When an instance of an RPC class (task or instrument) is created,
    QMI automatically detects any `QMI_Signal` declarations in the RPC class
    and converts them to `QMI_RegisteredSignal` instances in the RPC object instance.

    An instance of `QMI_RegisteredSignal` may be used to publish signals.
    """

    __slots__ = ("context",
                 "publisher_name",
                 "signal_name",
                 "arg_types")

    def __init__(self,
                 context: "qmi.core.context.QMI_Context",
                 publisher_name: str,
                 signal_name: str,
                 arg_types: Tuple[Type, ...]
                 ) -> None:
        """Initialize the `QMI_RegisteredSignal` instance.

        This is used internally by QMI.

        Do not create your own instances of `QMI_RegisteredSignal`.
        Instead, use the signal attributes of your QMI_RpcObject instance.
        """
        self.context = context
        self.publisher_name = publisher_name
        self.signal_name = signal_name
        self.arg_types = arg_types

    def __repr__(self) -> str:
        arg_types = ", ".join(arg_type.__name__ for arg_type in self.arg_types)
        return "<registered signal {}.{} ({})>".format(self.publisher_name, self.signal_name, arg_types)

    def publish(self, *args: Any) -> None:
        """Publish this signal.

        Arguments may be passed when publishing the signal. These must
        correspond to the argument types registered for this signal.

        The published signal will be received by all receivers currently
        subscribed to this signal type.

        This method is thread-safe: it may be called from any thread.
        """
        self.context.publish_signal(self.publisher_name, self.signal_name, *args)


class QMI_SignalSubscriber:
    """Represents a named signal, bound to a specific RPC proxy instance.

    When a proxy for an RPC class (task or instrument) is created,
    QMI automatically detects any `QMI_Signal` declarations in the RPC class
    and converts them to `QMI_SignalSubscriber` instances in the RPC proxy.

    An instance of `QMI_SignalSubscriber` may be used to subscribe to signals
    or unsubscribe from signals.
    """

    __slots__ = ("context",
                 "publisher_context",
                 "publisher_name",
                 "signal_name",
                 "signal_arg_types")

    def __init__(self,
                 context: "qmi.core.context.QMI_Context",
                 publisher_context: str,
                 publisher_name: str,
                 signal_name: str,
                 signal_arg_types: str
                 ) -> None:
        """Initialize the QMI_SignalSubscriber instance.

        This is used internally by QMI.

        Do not create your own instances of QMI_SignalSubscriber.
        Instead, use the signal attributes of an RPC proxy instance.
        """
        self.context = context
        self.publisher_context = publisher_context
        self.publisher_name = publisher_name
        self.signal_name = signal_name
        self.signal_arg_types = signal_arg_types

    def __repr__(self) -> str:
        return "<subscriber for signal {}.{}.{} {}>".format(self.publisher_context, self.publisher_name,
                                                            self.signal_name, self.signal_arg_types)

    def subscribe(self, receiver: "QMI_SignalReceiver") -> None:
        """Subscribe the specified receiver to this signal type.

        While subscribed, the `QMI_SignalReceiver` instance will receive and
        queue all published signals that match this subscription.

        A `QMI_SignalReceiver` instance can be simultaneously subscribed to
        multiple signals (either from the same publisher or from different publishers).
        Similarly, multiple receivers can be simultaneously subscribed to the
        same signal. However it is an error to try to subscribe a receiver to
        a signal to which it is already subscribed.

        Parameters:
            receiver: A `QMI_SignalReceiver` instance which will receive the published signals.
        """
        self.context.subscribe_signal(self.publisher_context, self.publisher_name, self.signal_name, receiver)

    def unsubscribe(self, receiver: "QMI_SignalReceiver") -> None:
        """Unsubscribe the specified receiver from this signal type.

        Parameters:
            receiver: The `QMI_SignalReceiver` instance to unsubscribe.
        """
        self.context.unsubscribe_signal(self.publisher_context, self.publisher_name, self.signal_name, receiver)


def _wait_for_condition(cond: threading.Condition, predicate: Callable[[], bool], timeout: Optional[float]) -> bool:
    """Helper function to wait for a condition.

    When called from the main thread, this function is equivalent to::

      cond.wait_for(predicate, timeout)

    When called from a task thread, this function uses a special method
    of waiting which stops waiting when the task is requested to stop.
    """
    import qmi.core.task
    thread = threading.current_thread()
    if isinstance(thread, qmi.core.task._TaskThread):
        # We are called from a task thread.
        # Use a special method to wait on the condition variable
        # which wakes up if the thread is requested to stop.
        return thread.wait_for_condition(cond, predicate, timeout)
    else:
        # Not called from a QMI_Thread.
        # We can use a normal wait on the condition variable.
        return cond.wait_for(predicate, timeout)


class QMI_SignalReceiver:
    """An instance of `QMI_SignalReceiver` contains a queue of received signals.

    A `QMI_SignalReceiver` instance can be subscribed to a specific set of
    signals. When any such signal gets published, the published signal is
    automatically added to the receive queue of the `QMI_SignalReceiver`.

    A task or script which creates a `QMI_SignalReceiver`, should periodically
    check to see if new signals have arrived and handle them.

    When an instance of `QMI_SignalReceiver` is no longer needed, it **must**
    be explicitly unsubscribed from all signals.
    """

    DISCARD_OLD = 1
    DISCARD_NEW = 2

    def __init__(self, max_queue_length: int = 10000, discard_policy: int = DISCARD_OLD) -> None:
        """Create a `QMI_SignalReceiver`, containing a signal receive queue.

        Parameters:
            max_queue_length: Maximum number of signals to keep in the receive queue.
            discard_policy: Policy in case the receive queue becomes full.
                If discard_policy == `QMI_SignalReceiver.DISCARD_OLD`, the oldest pending
                signal will be discarded to make room for a new publised signal.
                If discard_policy == `QMI_SignalReceiver.DISCARD_NEW`, newly received
                signals will be dropped when they do not fit in the queue.
        """
        assert max_queue_length > 0
        assert discard_policy in (self.DISCARD_OLD, self.DISCARD_NEW)
        self._max_queue_length = max_queue_length
        self._discard_policy = discard_policy
        self._queue = deque(maxlen=max_queue_length)  # type: deque
        self._queue_cond = threading.Condition()
        self._receiver_seqnr = 0

    def discard_all(self) -> None:
        """Discard all pending signals currently waiting in the receive queue.

        This method is thread-safe.
        """
        with self._queue_cond:
            self._queue.clear()

    def has_signal_ready(self) -> bool:
        """Return `True` if at least one received signal is waiting in the receive queue.

        This method is thread-safe.
        """
        with self._queue_cond:
            return len(self._queue) != 0

    def get_queue_length(self) -> int:
        """Return the number of signals currently in the receive queue.

        This method is thread-safe. Note however that the queue length can
        change at any time due to actions of other threads.
        """
        with self._queue_cond:
            return len(self._queue)

    def get_next_signal(self, timeout: Optional[float] = 0) -> ReceivedSignal:
        """Return the oldest published signal waiting in the receive queue.

        If there is no signal waiting in the queue, optionally wait until
        a new signal is received, subject to the specified timeout.

        This method is thread-safe.

        Parameters:
            timeout: Maximum time (in seconds) to wait for a new signal
                if the queue is empty, or None to wait indefinitely.

        Returns:
            `ReceivedSignal` tuple describing the received signal.

        Raises:
            QMI_TimeoutException: If the timeout expires before a signal is received.
            QMI_TaskStopException: If the calling task receives a stop request before a signal is received.
        """
        with self._queue_cond:
            if len(self._queue) == 0:
                predicate = lambda: (len(self._queue) > 0)
                if not _wait_for_condition(self._queue_cond, predicate, timeout):
                    raise QMI_TimeoutException("Timeout while waiting for signal")

            return self._queue.popleft()

    def _receive_signal(self, message: "QMI_SignalMessage") -> None:
        """Internal method to add a new received signal to the queue."""

        with self._queue_cond:
            sig = ReceivedSignal(
                publisher_context=message.source_address.context_id,
                publisher_name=message.source_address.object_id,
                signal_name=message.signal_name,
                args=message.args,
                receiver_seqnr=self._receiver_seqnr
            )
            self._receiver_seqnr += 1

            if len(self._queue) == self._max_queue_length:
                if self._discard_policy == self.DISCARD_NEW:
                    # Queue is full; ignore new signals.
                    return

            # Append new signal to queue.
            # If the queue is full, it will automatically discard the oldest signal.
            self._queue.append(sig)

            # Notify any thread waiting for a signal.
            self._queue_cond.notify_all()


class QMI_SignalMessage(QMI_Message):
    """Message sent to broadcast a signal between contexts.

    This class is intended for internal use within QMI. Application programs
    should not interact with this class directly.

    Attributes:
        signal_name: Name of the published signal.
            (The identity of the publisher follows from the `source_address` attribute.)
        args: Tuple of parameter values passed when publishing this signal.
    """

    __slots__ = ("signal_name", "args")

    def __init__(self,
                 source_address: QMI_MessageHandlerAddress,
                 destination_address: QMI_MessageHandlerAddress,
                 signal_name: str,
                 args: tuple
                 ) -> None:
        super().__init__(source_address, destination_address)
        self.signal_name = signal_name
        self.args = args


class QMI_SignalSubscriptionRequest(QMI_RequestMessage):
    """Message sent to subscribe or unsubscribe to a remote publisher.

    This message is sent to a remote `SignalManager` to subscribe or
    unsubscribe to the specified type of signal. The remote `SignalManager`
    will typically answer by sending a `QMI_SignalSubscriptionReply`.

    This class is intended for internal use within QMI. Application programs
    should not interact with this class directly.

    Attributes:
        publisher_name: Name of the RPC object that publishes the signal.
        signal_name: Signal name.
        subscribe: `True` to subscribe, `False` to unsubscribe.
    """

    __slots__ = ("publisher_name", "signal_name", "subscribe")

    def __init__(self,
                 source_address: QMI_MessageHandlerAddress,
                 destination_address: QMI_MessageHandlerAddress,
                 publisher_name: str,
                 signal_name: str,
                 subscribe: bool
                 ) -> None:
        super().__init__(source_address, destination_address)
        self.publisher_name = publisher_name
        self.signal_name = signal_name
        self.subscribe = subscribe


class QMI_SignalSubscriptionReply(QMI_ReplyMessage):
    """Message sent in response to a subscribe/unsubscribe request.

    This class is intended for internal use within QMI. Application programs
    should not interact with this class directly.

    Attributes:
        success: `True` if the request was processed successfully, `False` to indicate failure.
        error_msg: Short message describing the error, if `success` == `False`.
    """

    __slots__ = ("success", "error_msg")

    def __init__(self,
                 source_address: QMI_MessageHandlerAddress,
                 destination_address: QMI_MessageHandlerAddress,
                 request_id: str,
                 success: bool,
                 error_msg: str
                 ) -> None:
        super().__init__(source_address, destination_address, request_id)
        self.success = success
        self.error_msg = error_msg


class QMI_SignalRemovedMessage(QMI_Message):
    """Message sent to remote subscribers when an object stops publishing a signal.

    This message is currently sent only when the RPC object that was publishing
    the signal is removed from QMI. In this case, multiple instances of
    `QMI_SignalRemovedMessage` may be sent, one for each signal type that was
    published by the removed object.

    This class is intended for internal use within QMI. Application programs
    should not interact with this class directly.

    Attributes:
        publisher_name: Name of the RPC object name that was publishing the signal.
        signal_name: Name of the signal that is no longer published.
    """

    __slots__ = ("publisher_name", "signal_name")

    def __init__(self,
                 source_address: QMI_MessageHandlerAddress,
                 destination_address: QMI_MessageHandlerAddress,
                 publisher_name: str,
                 signal_name: str
                 ) -> None:
        super().__init__(source_address, destination_address)
        self.publisher_name = publisher_name
        self.signal_name = signal_name


class _PendingSubscriptionRequest:
    """Helper object representing a pending subscription request sent to a remote context."""

    def __init__(self,
                 publisher_context: str,
                 publisher_name: str,
                 signal_name: str,
                 subscribe: bool
                 ) -> None:
        """Initialize a new pending subscription request.

        Parameters:
            publisher_context:  Context ID of the remote context containing the publisher.
            publisher_name:     Object ID of the remote publisher.
            signal_name:        Signal name of the signal to subscribe to.
            subscribe:          True when subscribing, False when unsubscribing.
        """
        self.publisher_context = publisher_context
        self.publisher_name = publisher_name
        self.signal_name = signal_name
        self.subscribe = subscribe
        self.receivers: Set[QMI_SignalReceiver] = set()
        self._completed = threading.Event()
        self._success = False
        self._error_msg = ""

    def add_receiver(self, receiver: QMI_SignalReceiver) -> None:
        """Add a signal receiver to be added as local subscriber when the
        remote subscription request completes."""
        self.receivers.add(receiver)

    def set_reply(self, success: bool, error_msg: str) -> None:
        """Called when the reply is received.

        This will cause all current and future calls to the "wait()" method
        to complete and return the specified values.
        """
        assert not self._completed.is_set()
        self._success = success
        self._error_msg = error_msg
        self._completed.set()

    def wait(self) -> Tuple[bool, str]:
        """Wait until a reply is received, then return (success, error_msg)."""
        self._completed.wait()
        return (self._success, self._error_msg)


class SignalManager(QMI_MessageHandler):
    """Keeps track of signal subscriptions and handles signal publishing.

    Each context owns exactly one SignalManager.

    The SignalManager registers itself as a MessageHandler to receive messages
    sent to the ``"$pubsub"`` object of the local context.

    This class is intended for internal use within QMI. Application programs
    should not interact with this class directly.
    """

    PUBSUB_OBJECT_ID = "$pubsub"

    def __init__(self, context: "qmi.core.context.QMI_Context"):
        """Initialize the signal manager and register as message handler."""

        address = QMI_MessageHandlerAddress(context.name, self.PUBSUB_OBJECT_ID)
        super().__init__(address)

        _logger.debug("SignalManager initializing")

        self._context = context

        # Mutex to guard manipulation of internal data structures.
        self._lock = threading.Lock()

        # Register of subscribers in the local context.
        # Map "<context>.<publisher>.<signal>" to a set of SignalReceiver objects.
        self._local_subscriptions: Dict[str, Set[QMI_SignalReceiver]] = {}

        # Register of remote subscribers to locally published signals.
        # Type: { "publisher_name": { "signal_name": set_of_context_names } }
        # Map "<publisher>.<signal>" to a set of remote context names.
        self._remote_subscriptions: Dict[str, Set[str]] = {}

        # Register of pending subscription requests by request ID.
        # Each pending subscription request is stored in this dictionary,
        # as well as in "_pending_subscription_request_by_signal_name".
        self._pending_subscription_request_by_request_id: Dict[str, _PendingSubscriptionRequest] = {}

        # Register of pending subscription requests by signal name.
        # This is a mapping from "<context>.<publisher>.<signal>" to
        # a corresponding PendingSubscriptionRequest, if one exists.
        # Each pending subscription request is stored in this dictionary,
        # as well as in "_pending_subscription_request_by_request_id".
        self._pending_subscription_request_by_signal_name: Dict[str, _PendingSubscriptionRequest] = {}

        context.register_message_handler(self)

    def subscribe_signal(self,
                         publisher_context: str,
                         publisher_name: str,
                         signal_name: str,
                         receiver: QMI_SignalReceiver) -> None:
        """Subscribe a SignalReceiver to a specified signal.

        This method blocks until the subscription is established.
        This method is thread-safe and may safely be called from any thread.

        Raises:
            QMI_UnknownNameException: When the specified publisher does not exist.
            QMI_MessageDeliveryException: When the subscription request can not be routed to a remote context.
        """

        if publisher_context == "":
            publisher_context = self._context.name
        if not is_valid_object_name(publisher_context):
            raise QMI_UsageException("Invalid context name {!r}".format(publisher_context))
        if not is_valid_object_name(publisher_name):
            raise QMI_UsageException("Invalid publisher name {!r}".format(publisher_name))
        if not is_valid_object_name(signal_name):
            raise QMI_UsageException("Invalid signal name {!r}".format(signal_name))

        _logger.debug("Subscribing to signal %s.%s.%s", publisher_context, publisher_name, signal_name)

        if publisher_context == self._context.name:
            self._subscribe_local(publisher_context, publisher_name, signal_name, receiver)
        else:
            self._subscribe_remote(publisher_context, publisher_name, signal_name, receiver)

    def unsubscribe_signal(self,
                           publisher_context: str,
                           publisher_name: str,
                           signal_name: str,
                           receiver: QMI_SignalReceiver) -> None:
        """Unsubscribe a SignalReceiver from a specified signal.

        If the receiver is not currently subscribed to the specified signal,
        this function will do nothing.

        This method is thread-safe and may safely be called from any thread.
        """

        if publisher_context == "":
            publisher_context = self._context.name
        if not is_valid_object_name(publisher_context):
            raise QMI_UsageException("Invalid context name {!r}".format(publisher_context))
        if not is_valid_object_name(publisher_name):
            raise QMI_UsageException("Invalid publisher name {!r}".format(publisher_name))
        if not is_valid_object_name(signal_name):
            raise QMI_UsageException("Invalid signal name {!r}".format(signal_name))

        _logger.debug("Unsubscribing from signal %s.%s.%s", publisher_context, publisher_name, signal_name)

        if publisher_context == self._context.name:
            self._remove_local_subscriber(publisher_context, publisher_name, signal_name, receiver)
        else:
            self._unsubscribe_remote(publisher_context, publisher_name, signal_name, receiver)

    def _add_local_subscriber(self,
                              publisher_context: str,
                              publisher_name: str,
                              signal_name: str,
                              receiver: QMI_SignalReceiver) -> None:
        """Add the receiver to the list of local subscribers."""
        full_name = publisher_context + "." + publisher_name + "." + signal_name
        with self._lock:
            lsubs = self._local_subscriptions.get(full_name)
            if lsubs is None:
                lsubs = set()
                self._local_subscriptions[full_name] = lsubs
            lsubs.add(receiver)

    def _remove_local_subscriber(self,
                                 publisher_context: str,
                                 publisher_name: str,
                                 signal_name: str,
                                 receiver: QMI_SignalReceiver) -> None:
        """Remove the receiver from the list of local subscribers."""
        full_name = publisher_context + "." + publisher_name + "." + signal_name
        with self._lock:
            lsubs = self._local_subscriptions.get(full_name)
            if lsubs is not None:
                lsubs.discard(receiver)
                if len(lsubs) == 0:
                    self._local_subscriptions.pop(full_name)

    def _add_remote_subscriber(self, publisher_name: str, signal_name: str, subscriber_context: str) -> None:
        full_name = publisher_name + "." + signal_name
        with self._lock:
            rsubs = self._remote_subscriptions.get(full_name)
            if rsubs is None:
                rsubs = set()
                self._remote_subscriptions[full_name] = rsubs
            rsubs.add(subscriber_context)

    def _remove_remote_subscriber(self, publisher_name: str, signal_name: str, subscriber_context: str) -> None:
        full_name = publisher_name + "." + signal_name
        with self._lock:
            rsubs = self._remote_subscriptions.get(full_name)
            if rsubs is not None:
                rsubs.discard(subscriber_context)
                if len(rsubs) == 0:
                    self._remote_subscriptions.pop(full_name)

    def _subscribe_local(self,
                         publisher_context: str,
                         publisher_name: str,
                         signal_name: str,
                         receiver: QMI_SignalReceiver
                         ) -> None:
        """Subscribe to a signal from a local publisher."""

        assert publisher_context == self._context.name

        # Check that the publisher exists as an RPC object.
        if self._context.get_rpc_object_descriptor(publisher_name) is None:
            raise QMI_SignalSubscriptionException("Unknown RPC object {}.{}".format(publisher_context, publisher_name))

        # Add the receiver to the list of local subscribers.
        self._add_local_subscriber(publisher_context, publisher_name, signal_name, receiver)

        # Check that the publisher still exists.
        if self._context.get_rpc_object_descriptor(publisher_name) is None:
            # Publisher vanished just before or after we subscribed. Undo the subscription.
            self._remove_local_subscriber(publisher_context, publisher_name, signal_name, receiver)

    def _send_subscription_request(self, request_message: QMI_SignalSubscriptionRequest) -> None:
        """Send a subscription request message to the remote context."""
        _logger.debug("Sending %s request for %s.%s.%s",
                      ("subscribe" if request_message.subscribe else "unsubscribe"),
                      request_message.destination_address.context_id,
                      request_message.publisher_name,
                      request_message.signal_name)
        try:
            self._context.send_message(request_message)
        except QMI_MessageDeliveryException as exc:
            # Failed to send the subscription request.
            # Handle this as if we received an error reply.
            self._handle_subscription_reply(request_message.request_id, False, str(exc))

    def _subscribe_remote(self,
                         publisher_context: str,
                         publisher_name: str,
                         signal_name: str,
                         receiver: QMI_SignalReceiver
                         ) -> None:
        """Subscribe to a signal from a remote publisher."""

        full_name = publisher_context + "." + publisher_name + "." + signal_name
        request_message = None

        with self._lock:

            # Check if there are other local subscribers for the same signal.
            lsubs = self._local_subscriptions.get(full_name)
            if lsubs:
                # There already are some local subscribers.
                # Just add this receiver to the list and we are subscribed.
                lsubs.add(receiver)
                return

            # Otherwise, check if there is a pending subscription request for this signal.
            pending_request = self._pending_subscription_request_by_signal_name.get(full_name)
            if pending_request is None:
                # Create a subscription request message.
                request_message = QMI_SignalSubscriptionRequest(
                    source_address=QMI_MessageHandlerAddress(self._context.name, self.PUBSUB_OBJECT_ID),
                    destination_address=QMI_MessageHandlerAddress(publisher_context, self.PUBSUB_OBJECT_ID),
                    publisher_name=publisher_name,
                    signal_name=signal_name,
                    subscribe=True)
                # Add the new request to the pending subscription table.
                pending_request = _PendingSubscriptionRequest(publisher_context, publisher_name, signal_name, True)
                self._pending_subscription_request_by_signal_name[full_name] = pending_request
                self._pending_subscription_request_by_request_id[request_message.request_id] = pending_request

            # Add this receiver to the pending subscription request.
            pending_request.add_receiver(receiver)

        # Send the remote subscription request, if needed.
        if request_message is not None:
            self._send_subscription_request(request_message)

        # Wait until the pending subscription request completes.
        (success, error_msg) = pending_request.wait()
        if not success:
            raise QMI_SignalSubscriptionException(error_msg)

    def _unsubscribe_remote(self,
                            publisher_context: str,
                            publisher_name: str,
                            signal_name: str,
                            receiver: QMI_SignalReceiver) -> None:
        """Unsubscribe a receiver from a signal published by a remote object."""

        full_name = publisher_context + "." + publisher_name + "." + signal_name
        request_message = None

        with self._lock:

            # Remove the local subscription on the specified signal.
            last_subscriber = False
            lsubs = self._local_subscriptions.get(full_name)
            if lsubs is not None:
                lsubs.discard(receiver)
                if len(lsubs) == 0:
                    self._local_subscriptions.pop(full_name)
                    last_subscriber = True

            if last_subscriber:
                # We have just removed the last remaining local subscriber.
                # Create a remote unsubscribe request (unless there is already
                # a remote subscription request in progress.)
                pending_request = self._pending_subscription_request_by_signal_name.get(full_name)
                if pending_request is None:
                    # Create an unsubscribe request message.
                    request_message = QMI_SignalSubscriptionRequest(
                        source_address=QMI_MessageHandlerAddress(self._context.name, self.PUBSUB_OBJECT_ID),
                        destination_address=QMI_MessageHandlerAddress(publisher_context, self.PUBSUB_OBJECT_ID),
                        publisher_name=publisher_name,
                        signal_name=signal_name,
                        subscribe=False)

                    # Add the new request to the pending subscription table.
                    pending_request = _PendingSubscriptionRequest(publisher_context,
                                                                  publisher_name,
                                                                  signal_name,
                                                                  False)
                    self._pending_subscription_request_by_signal_name[full_name] = pending_request
                    self._pending_subscription_request_by_request_id[request_message.request_id] = pending_request

        # Send the remote unsubscribe request, if needed.
        if request_message is not None:
            self._send_subscription_request(request_message)

    def _deliver_local(self, message: "QMI_SignalMessage") -> None:
        """Deliver a published message to local subscribers."""

        assert message.destination_address.context_id == self._context.name
        assert message.destination_address.object_id == self.PUBSUB_OBJECT_ID

        full_name = (message.source_address.context_id
                     + "." + message.source_address.object_id
                     + "." + message.signal_name)

        with self._lock:
            receiver_set = self._local_subscriptions.get(full_name)
            if receiver_set is None:
                receiver_list: List[QMI_SignalReceiver] = []
            else:
                receiver_list = list(receiver_set)

        for receiver in receiver_list:
            receiver._receive_signal(message)

    def publish_signal(self, publisher_name: str, signal_name: str, args: tuple) -> None:
        """Publish the specified signal to the QMI network."""

        if not is_valid_object_name(publisher_name):
            raise QMI_UsageException("Invalid publisher name {!r}".format(publisher_name))
        if not is_valid_object_name(signal_name):
            raise QMI_UsageException("Invalid signal name {!r}".format(signal_name))

        source_address = QMI_MessageHandlerAddress(self._context.name, publisher_name)

        # Local delivery of locally published signal.
        msg = QMI_SignalMessage(
            source_address=source_address,
            destination_address=QMI_MessageHandlerAddress(self._context.name, self.PUBSUB_OBJECT_ID),
            signal_name=signal_name,
            args=args
        )
        self._deliver_local(msg)

        # Remote delivery of locally published signal.
        full_name = publisher_name + "." + signal_name

        with self._lock:
            rsubs = self._remote_subscriptions.get(full_name)
            if rsubs is None:
                rsubs_list: List[str] = []
            else:
                # Copy list of subscribers to avoid race conditions.
                rsubs_list = list(rsubs)

        for sub in rsubs_list:
            msg = QMI_SignalMessage(
                source_address=source_address,
                destination_address=QMI_MessageHandlerAddress(sub, self.PUBSUB_OBJECT_ID),
                signal_name=signal_name,
                args=args
            )
            try:
                self._context.send_message(msg)
            except QMI_MessageDeliveryException:
                # Signal could not be delivered to remote context - ignore.
                _logger.debug("Can not send signal to remote context %s", sub, exc_info=True)

    def _handle_subscription_request(self, request_message: QMI_SignalSubscriptionRequest) -> None:
        """Called when we receive a subscribe/unsubscribe request from a remote subscriber.

        This function creates or removes an entry in the table of remote subscribers,
        then sends a reply message.
        """
        if request_message.source_address.object_id != self.PUBSUB_OBJECT_ID:
            raise QMI_RuntimeException("Unexpected SignalSubscriptionRequest from {}.{}"
                                       .format(request_message.source_address.context_id,
                                               request_message.source_address.object_id))

        publisher_name = request_message.publisher_name
        signal_name = request_message.signal_name
        subscriber_context = request_message.source_address.context_id

        _logger.debug("Got %s request from %s for %s.%s",
                      ("subscribe" if request_message.subscribe else "unsubscribe"),
                      subscriber_context, publisher_name, signal_name)

        if request_message.subscribe:
            # Check that the publisher exists as a local RPC object.
            if self._context.get_rpc_object_descriptor(publisher_name) is None:
                success = False
                error_msg = "Unknown RPC object {}.{}".format(self._context.name, publisher_name)
            else:
                # Add the remote context to the table of remote subscribers.
                self._add_remote_subscriber(publisher_name, signal_name, subscriber_context)

                # Double-check that the publisher still exists.
                if self._context.get_rpc_object_descriptor(publisher_name) is None:
                    self._remove_remote_subscriber(publisher_name, signal_name, subscriber_context)
                    success = False
                    error_msg = "Unknown RPC object {}.{}".format(self._context.name, publisher_name)
                else:
                    success = True
                    error_msg = ""
        else:
            # Unsubscribe. This is always successful.
            self._remove_remote_subscriber(publisher_name, signal_name, subscriber_context)
            success = True
            error_msg = ""

        # Send reply message.
        reply_message = QMI_SignalSubscriptionReply(
            source_address=request_message.destination_address,
            destination_address=request_message.source_address,
            request_id=request_message.request_id,
            success=success,
            error_msg=error_msg)
        try:
            self._context.send_message(reply_message)
        except QMI_MessageDeliveryException:
            # Ignore errors during reply delivery.
            pass

    def _handle_subscription_reply(self, request_id: str, success: bool, error_msg: str) -> None:
        """Called when we receive a reply to a pending subscribe/unsubscribe request.

        This function clears the pending subscription request and notifies any
        waiting threads that the subscription has completed.

        A special case occurs when an unsubscribe request has completed while
        there are already new subscribers waiting for the same signal. This
        is handled by immediately sending a new subscribe request.
        """
        with self._lock:
            # Find the pending subscription request and remove it from the table.
            pending_request = self._pending_subscription_request_by_request_id.pop(request_id)
            full_name = (pending_request.publisher_context
                         + "." + pending_request.publisher_name
                         + "." + pending_request.signal_name)
            self._pending_subscription_request_by_signal_name.pop(full_name)

            _logger.debug("Got reply to %s request for %s, status %s",
                          ("subscribe" if pending_request.subscribe else "unsubscribe"),
                          full_name, success)

            # On successful completion of a subscribe request, move the waiting
            # subscribers to the list of local subscribers for this signal.
            if pending_request.subscribe and success:
                lsubs = self._local_subscriptions.get(full_name)
                if lsubs:
                    lsubs.update(pending_request.receivers)
                else:
                    self._local_subscriptions[full_name] = pending_request.receivers

            # When a subscribe request completes, notify all waiting subscribers.
            if pending_request.subscribe:
                pending_request.set_reply(success, error_msg)

            request_message = None
            if (not pending_request.subscribe) and pending_request.receivers:
                # An unsubscribe request just completed, but there are already
                # new subscribers waiting for the same signal.
                # Immediately send a new subscription request.
                request_message = QMI_SignalSubscriptionRequest(
                    source_address=QMI_MessageHandlerAddress(self._context.name, self.PUBSUB_OBJECT_ID),
                    destination_address=QMI_MessageHandlerAddress(pending_request.publisher_context,
                                                                  self.PUBSUB_OBJECT_ID),
                    publisher_name=pending_request.publisher_name,
                    signal_name=pending_request.signal_name,
                    subscribe=True)
                # Add the new request to the pending subscription table.
                pending_request.subscribe = True
                self._pending_subscription_request_by_signal_name[full_name] = pending_request
                self._pending_subscription_request_by_request_id[request_message.request_id] = pending_request

        # Send the remote subscription request, if needed.
        if request_message is not None:
            self._send_subscription_request(request_message)

    def _handle_remote_signal_removed(self, message: QMI_SignalRemovedMessage) -> None:
        """Called when we receive a notification that a remote publisher has been removed.

        This function removes all local subscriptions on the removed signal.
        """
        publisher_context = message.source_address.context_id
        publisher_name = message.publisher_name
        signal_name = message.signal_name
        full_name = publisher_context + "." + publisher_name + "." + signal_name
        with self._lock:
            if full_name in self._local_subscriptions:
                self._local_subscriptions.pop(full_name)

    def handle_message(self, message: QMI_Message) -> None:
        """Handle messages sent to the signal manager object."""

        assert message.destination_address.context_id == self._context.name
        assert message.destination_address.object_id == self.PUBSUB_OBJECT_ID

        if ((not isinstance(message, QMI_SignalMessage))
                and (message.source_address.object_id != self.PUBSUB_OBJECT_ID)):
            raise QMI_RuntimeException("Unexpected message {} from {}.{}"
                                       .format(type(message).__name__,
                                               message.source_address.context_id,
                                               message.source_address.object_id))

        if isinstance(message, QMI_SignalMessage):
            self._deliver_local(message)
        elif isinstance(message, QMI_SignalSubscriptionRequest):
            self._handle_subscription_request(message)
        elif isinstance(message, QMI_SignalSubscriptionReply):
            self._handle_subscription_reply(message.request_id, message.success, message.error_msg)
        elif isinstance(message, QMI_ErrorReplyMessage):
            self._handle_subscription_reply(message.request_id, False, message.error_msg)
        elif isinstance(message, QMI_SignalRemovedMessage):
            self._handle_remote_signal_removed(message)
        else:
            raise QMI_RuntimeException("Unexpected message type {}".format(type(message).__name__))

    def handle_object_removed(self, rpc_object_name: str) -> None:
        """Called when a local RPC object is removed.

        This function drops any local subscriptions on signals published by
        the removed object. This function also drops any remote subscribers
        on signals published by the removed object, and notifies the remote
        subscribers that the signal has been removed.
        """
        with self._lock:
            # Drop any local subscriptions on signals published by the removed object.
            # This is potentially slow.
            pattern = self._context.name + "." + rpc_object_name + "."
            remove_entries = []
            for full_name in self._local_subscriptions:
                if full_name.startswith(pattern):
                    remove_entries.append(full_name)
            for full_name in remove_entries:
                self._local_subscriptions.pop(full_name)

            # Drop any remote subscribers in the removed peer context.
            # This is potentially slow.
            pattern = rpc_object_name + "."
            remove_entries = []
            notify_subscribers = []
            for full_name in self._remote_subscriptions:
                if full_name.startswith(pattern):
                    remove_entries.append(full_name)
            for full_name in remove_entries:
                rsubs = self._remote_subscriptions.pop(full_name)
                (_publisher_name, signal_name) = full_name.split(".")
                for subscriber_context in rsubs:
                    notify_subscribers.append((signal_name, subscriber_context))

        # Notify remote subscribers that these signals have been removed.
        for (signal_name, subscriber_context) in notify_subscribers:
            _logger.debug("Sending signal removed notification to %s for %s.%s",
                          subscriber_context, rpc_object_name, signal_name)
            message = QMI_SignalRemovedMessage(
                source_address=QMI_MessageHandlerAddress(self._context.name, self.PUBSUB_OBJECT_ID),
                destination_address=QMI_MessageHandlerAddress(subscriber_context, self.PUBSUB_OBJECT_ID),
                publisher_name=rpc_object_name,
                signal_name=signal_name)
            try:
                self._context.send_message(message)
            except QMI_MessageDeliveryException as exc:
                # Ignore errors during message delivery.
                pass

    def handle_peer_context_removed(self, context_name: str) -> None:
        """Update remote signal administration after disconnecting a peer context.

        This method is called by the socket manager when a peer connection
        (outgoing or incoming) is closed. This method drops any remote subscribers
        in the removed context. This method also drops any local subscriptions
        on signals published by the removed context.
        """
        with self._lock:
            # Drop any remote subscribers in the removed peer context.
            # This is potentially slow, but disconnecting from a peer context should not occur very frequently.
            remove_entries = []
            for (full_name, rsubs) in self._remote_subscriptions.items():
                if context_name in rsubs:
                    rsubs.remove(context_name)
                    if len(rsubs) == 0:
                        remove_entries.append(full_name)
            for full_name in remove_entries:
                self._remote_subscriptions.pop(full_name)

            # Drop any local subscriptions on signals published by the remote context.
            # This is potentially slow.
            pattern = context_name + "."
            remove_entries = []
            for full_name in self._local_subscriptions:
                if full_name.startswith(pattern):
                    remove_entries.append(full_name)
            for full_name in remove_entries:
                self._local_subscriptions.pop(full_name)


# Imports needed only for static typing.
if typing.TYPE_CHECKING:
    import qmi.core.context



"""
Boston Micromachines, deformable mirrors.

The qmi.instruments.boston_micromachines package supports Multi-DM models with 140 actuators.
"""

from qmi.instruments.boston_micromachines.multidm import BostonMicromachines_MultiDM
from qmi.instruments.boston_micromachines.multidm import write_mirror_shape, read_mirror_shape



import tests.instruments.picotech.ps3000a_stub as ps3000a
import tests.instruments.picotech.ps4000a_stub as ps4000a

class picosdk:
    def __init__(self):
        self.ps3000a = ps3000a.ps3000a()
        self.ps4000a = ps4000a.ps4000a()



"""Actuators for the single axis motion controllers."""
from typing import NamedTuple


class TravelRange(NamedTuple):
    """Travel range """
    min: float
    max: float


class LinearActuator:
    """Linear actuator."""

    def __init__(self,
                 travel_range_min: float,
                 travel_range_max: float,
                 max_velocity: float,
                 min_velocity: float,
                 min_incremental_motion: float,
                 push_force: float,
                 encoder_resolution: float
                 ) -> None:
        """Initialize actuator.

        Parameters:
            travel_range_min:           Minimum allowed range for actuator in mm.
            travel_range_max:           Maximum allowed range for actuator in mm.
            max_velocity:               Maximum velocity in mm/s.
            min_velocity:               Minimum velocity in mm/s.
            min_incremental_motion:     Minimum incremental motion in mm.
            push_force:                 Maximum push force in N.
            encoder_resolution:         Resolution of encoder in mm
        """
        self.TRAVEL_RANGE = TravelRange(travel_range_min, travel_range_max)
        self.MAX_VELOCITY = max_velocity
        self.MIN_VELOCITY = min_velocity
        self.MIN_INCREMENTAL_MOTION = min_incremental_motion
        self.PUSH_FORCE = push_force
        self.ENCODER_RESOLUTION = encoder_resolution


# max speed set to 0.03mm/s as axials loads over 45N can only support 0.03mm/s
TRA12CC: LinearActuator = LinearActuator(
    0, 12, 0.03, 0.000001, 0.0002, 60, 0.000030517578)
TRB6CC: LinearActuator = LinearActuator(
    0, 6, 0.2, 0.000001, 0.0001, 90, 0.00001447)

CMA25CCL: LinearActuator = LinearActuator(
    0, 25, 0.4, 0.05, 0.0002, 90, 0.000048828)

UTS100PP: LinearActuator = LinearActuator(
    -50, 50, 20, 0.05, 300e-6, 200, 0.0)



#! /usr/bin/env python3

"""
Test program for QMI_SerialTransport.

This program needs two serial ports (or USB serial port adapters)
which are connected by an RS-232 cross cable.
"""

import argparse
import queue
import threading
import time

import numpy as np
import serial

import qmi
from qmi.core.exceptions import QMI_TimeoutException
from qmi.core.transport import create_transport


def _make_message(message_length):
    return bytes([ord("a") + (i % 26) for i in range(message_length)])


def _report_interval_stats(text, data):
    print("  {:40}: avg={:.4f}, min={:.4f}, max={:.4f} (s)"
          .format(text, np.mean(data), np.min(data), np.max(data)))


def _test_write_helper(serial_bench, message_lengths, recv_messages, recv_timestamps):
    """Helper function to receive serial port data in a background thread."""

    for message_length in message_lengths:

        # Receive message and record timestamp.
        msg = serial_bench.read(message_length)
        t = time.monotonic()

        recv_messages.append(msg)
        recv_timestamps.append(t)

        if len(msg) != message_length:
            break


def test_write(transport, serial_bench):
    """Test writing through transport."""

    print()
    print("Test writing through transport:")

    message_lengths = [1, 2, 3, 5, 10, 15, 20, 25, 30, 35]
    send_messages = []
    send_timestamps = []
    recv_messages = []
    recv_timestamps = []

    thread = threading.Thread(target=_test_write_helper, kwargs={
        "serial_bench": serial_bench,
        "message_lengths": message_lengths,
        "recv_messages": recv_messages,
        "recv_timestamps": recv_timestamps
    })
    thread.start()

    try:
        for message_length in message_lengths:

            # Wait to decouple from previous message.
            time.sleep(0.1)

            # Create test message.
            msg = _make_message(message_length)
            send_messages.append(msg)

            # Record timestamp, then write message.
            t = time.monotonic()
            transport.write(msg)
            send_timestamps.append(t)

    finally:
        # Wait for background thread to end.
        thread.join()

    if len(recv_timestamps) != len(send_timestamps):
        print("ERROR: expected {} messages but received only {}"
              .format(len(send_timestamps), len(recv_timestamps)))
        return

    for (send_msg, recv_msg) in zip(send_messages, recv_messages):
        if recv_msg != send_msg:
            print("ERROR: expected message {!r} but received {!r}"
                  .format(send_msg, recv_msg))
            return

    # Report send-to-receive latency.
    wait_times = np.array(recv_timestamps) - np.array(send_timestamps)
    _report_interval_stats("send-to-receive latency", wait_times)


def _test_read_helper(serial_bench, message_queue):
    """Helper function to send data through the serial port in a background thread."""

    while True:

        # Get next message.
        (msg, msg_delay) = message_queue.get(block=True)

        if msg is None:
            break

        # Wait before starting to send the message.
        time.sleep(msg_delay)

        # Send the message.
        serial_bench.write(msg)


def _test_read_message_ready(transport, serial_bench, timeout):
    """Test reading a message which is already available."""

    read_latency = []
    for message_length in [1, 2, 3, 5, 10, 15, 20, 25, 30, 35]:

        # Send message through test bench.
        msg = _make_message(message_length)
        serial_bench.write(msg)

        # Wait to make sure message is received.
        time.sleep(0.2)

        # Read message and measure latency.
        t0 = time.monotonic()
        recv_msg = transport.read(nbytes=message_length, timeout=timeout)
        t1 = time.monotonic()
        read_latency.append(t1 - t0)

        if recv_msg != msg:
            print("ERROR: Expecting {!r} but received {!r}".format(msg, recv_msg))

    _report_interval_stats("read latency, message ready", read_latency)


def _test_read_no_data(transport, timeout):
    """Test timeout when nothing is received."""

    read_latency = []
    for message_length in [1, 2, 3, 5, 10, 15, 20, 25, 30, 35]:

        # Sleep to decouple from previous message.
        time.sleep(0.1)

        # Read message and measure latency until timeout.
        try:
            t0 = time.monotonic()
            recv_msg = transport.read(nbytes=message_length, timeout=timeout)
            print("ERROR: Expecting timeout but received {!r}".format(recv_msg))
        except QMI_TimeoutException:
            t1 = time.monotonic()
        read_latency.append(t1 - t0)

    _report_interval_stats("timeout latency, no data", read_latency)


def _test_read_partial_data(transport, serial_bench, timeout):
    """Test timeout when a partial message is received."""

    read_latency = []
    for message_length in [2, 3, 5, 10, 15, 20, 25, 30, 35]:

        # Send partial message.
        msg = _make_message(message_length // 2)
        serial_bench.write(msg)

        # Sleep to make sure partial message is available.
        time.sleep(0.2)

        # Read message and measure latency until timeout.
        try:
            t0 = time.monotonic()
            recv_msg = transport.read(nbytes=message_length, timeout=timeout)
            print("ERROR: Expecting timeout but received {!r}".format(recv_msg))
        except QMI_TimeoutException:
            t1 = time.monotonic()
        read_latency.append(t1 - t0)

        # Read partial message to flush it from the buffer.
        recv_msg = transport.read(nbytes=len(msg), timeout=0.0)
        if recv_msg != msg:
            print("ERROR: Expecting {!r} but received {!r}".format(msg, recv_msg))

    _report_interval_stats("timeout latency, partial data", read_latency)


def _test_read_wait_for_message(transport, message_queue, timeout, fragments):
    """Test latency when a message is received after waiting."""

    assert fragments in (1, 2)

    msg_delay = 0.5 if (timeout is None) else (0.4 * timeout)

    read_latency = []
    for message_length in range(fragments, 10+fragments):

        # Sleep to decouple from previous message.
        time.sleep(0.1)

        # Tell background thread to send message (in one or two fragments):
        msg = _make_message(message_length)
        if fragments == 2:
            message_queue.put((msg[:message_length//2], 0.5 * msg_delay))
            message_queue.put((msg[message_length//2:], 0.5 * msg_delay))
        else:
            message_queue.put((msg, msg_delay))

        # Read message and measure latency.
        t0 = time.monotonic()
        recv_msg = transport.read(nbytes=message_length, timeout=timeout)
        t1 = time.monotonic()
        read_latency.append(t1 - t0)

        if recv_msg != msg:
            print("ERROR: Expecting {!r} but received {!r}".format(msg, recv_msg))

    if fragments == 2:
        _report_interval_stats("read latency, 2*{:.3f}s message delay".format(0.5 * msg_delay), read_latency)
    else:
        _report_interval_stats("read latency, {:.3f}s message delay".format(msg_delay), read_latency)


def test_read(transport, serial_bench, timeout):
    """Test reading through transport."""

    print()
    print("Test reading fixed number of bytes through transport (timeout={}):".format(timeout))

    _test_read_message_ready(transport, serial_bench, timeout)
    if timeout is not None:
        _test_read_no_data(transport, timeout)
        _test_read_partial_data(transport, serial_bench, timeout)

    if (timeout is None) or (timeout > 0):

        # Start backgroud thread.
        message_queue = queue.Queue()
        thread = threading.Thread(target=_test_read_helper, kwargs={
            "serial_bench": serial_bench,
            "message_queue": message_queue
        })
        thread.start()

        try:
            _test_read_wait_for_message(transport, message_queue, timeout, 1)
            _test_read_wait_for_message(transport, message_queue, timeout, 2)

        finally:
            # Wait for background thread to end.
            message_queue.put((None, 0.0))
            thread.join()


def _test_read_until_message_ready(transport, serial_bench, message_terminator, timeout):
    """Test reading a message which is already available."""

    read_latency = []
    for message_length in [1, 2, 3, 5, 10, 15, 20, 25, 30, 35]:

        # Send message through test bench.
        msg = _make_message(message_length) + message_terminator
        serial_bench.write(msg)

        # Wait to make sure message is received.
        time.sleep(0.2)

        # Read message and measure latency.
        t0 = time.monotonic()
        recv_msg = transport.read_until(message_terminator=message_terminator, timeout=timeout)
        t1 = time.monotonic()
        read_latency.append(t1 - t0)

        if recv_msg != msg:
            print("ERROR: Expecting {!r} but received {!r}".format(msg, recv_msg))

    _report_interval_stats("read latency, message ready", read_latency)


def _test_read_until_no_data(transport, message_terminator, timeout):
    """Test timeout when nothing is received."""

    read_latency = []
    for i in range(10):

        # Sleep to decouple from previous message.
        time.sleep(0.1)

        # Read message and measure latency until timeout.
        try:
            t0 = time.monotonic()
            recv_msg = transport.read_until(message_terminator=message_terminator, timeout=timeout)
            print("ERROR: Expecting timeout but received {!r}".format(recv_msg))
        except QMI_TimeoutException:
            t1 = time.monotonic()
        read_latency.append(t1 - t0)

    _report_interval_stats("timeout latency, no data", read_latency)


def _test_read_until_partial_data(transport, serial_bench, message_terminator, timeout):
    """Test timeout when a partial message is received."""

    read_latency = []
    for message_length in [1, 2, 3, 5, 8, 10, 12, 15, 20]:

        # Send partial message.
        msg = _make_message(message_length)
        serial_bench.write(msg)

        # Sleep to make sure partial message is available.
        time.sleep(0.2)

        # Read message and measure latency until timeout.
        try:
            t0 = time.monotonic()
            recv_msg = transport.read_until(message_terminator=message_terminator, timeout=timeout)
            print("ERROR: Expecting timeout but received {!r}".format(recv_msg))
        except QMI_TimeoutException:
            t1 = time.monotonic()
        read_latency.append(t1 - t0)

        # Read partial message to flush it from the buffer.
        recv_msg = transport.read(nbytes=message_length, timeout=0.0)
        if recv_msg != msg:
            print("ERROR: Expecting {!r} but received {!r}".format(msg, recv_msg))

    _report_interval_stats("timeout latency, partial data", read_latency)


def _test_read_until_wait_for_message(transport, message_queue, message_terminator, timeout, fragments):
    """Test latency when a message is received after waiting."""

    assert fragments in (1, 2)

    msg_delay = 0.5 if (timeout is None) else (0.4 * timeout)

    read_latency = []
    for message_length in range(fragments, 10+fragments):

        # Sleep to decouple from previous message.
        time.sleep(0.1)

        # Tell background thread to send message (in one or two fragments):
        msg = _make_message(message_length) + message_terminator
        if fragments == 2:
            message_queue.put((msg[:message_length//2], 0.5 * msg_delay))
            message_queue.put((msg[message_length//2:], 0.5 * msg_delay))
        else:
            message_queue.put((msg, msg_delay))

        # Read message and measure latency.
        t0 = time.monotonic()
        recv_msg = transport.read_until(message_terminator=message_terminator, timeout=timeout)
        t1 = time.monotonic()
        read_latency.append(t1 - t0)

        if recv_msg != msg:
            print("ERROR: Expecting {!r} but received {!r}".format(msg, recv_msg))

    if fragments == 2:
        _report_interval_stats("read latency, 2*{:.3f}s message delay".format(0.5 * msg_delay), read_latency)
    else:
        _report_interval_stats("read latency, {:.3f}s message delay".format(msg_delay), read_latency)


def test_read_until(transport, serial_bench, message_terminator, timeout):
    """Test reading through transport until end of message."""

    print()
    print("Test reading through transport until {!r} (timeout={}):".format(message_terminator, timeout))

    _test_read_until_message_ready(transport, serial_bench, message_terminator, timeout)
    if timeout is not None:
        _test_read_until_no_data(transport, message_terminator, timeout)
        _test_read_until_partial_data(transport, serial_bench, message_terminator, timeout)

    if (timeout is None) or (timeout > 0):

        # Start backgroud thread.
        message_queue = queue.Queue()
        thread = threading.Thread(target=_test_read_helper, kwargs={
            "serial_bench": serial_bench,
            "message_queue": message_queue
        })
        thread.start()

        try:
            _test_read_until_wait_for_message(transport, message_queue, message_terminator, timeout, 1)
            _test_read_until_wait_for_message(transport, message_queue, message_terminator, timeout, 2)

        finally:
            # Wait for background thread to end.
            message_queue.put((None, 0.0))
            thread.join()


def test_serial_transport(dutdev, benchdev, baudrate):
    """Test serial port transport at a specific baudrate."""

    print()
    print("Testing at baud rate {}".format(baudrate))
    print()

    print("Opening test bench serial port {} ...".format(benchdev))
    serial_bench = serial.Serial(benchdev, baudrate, timeout=5.0)
    print("  ok")

    try:

        print("Opening serial transport device {} ...".format(dutdev))
        transport = create_transport("serial:{}:baudrate={}".format(dutdev, baudrate))
        transport.open()
        print("  ok")

        try:
            test_write(transport, serial_bench)
            for timeout in (None, 0.0, 0.1, 1.0):
                test_read(transport, serial_bench, timeout)
            for timeout in (None, 0.0, 0.5):
                test_read_until(transport, serial_bench, b"\n", timeout)
                test_read_until(transport, serial_bench, b"\r\n", timeout)

        finally:
            transport.close()

    finally:
        serial_bench.close()


def main():

    parser = argparse.ArgumentParser()
    parser.description = "Test program for QMI_SerialTransport."
    parser.add_argument("dutdev", action="store", type=str,
                        help="serial port device under test (e.g. /dev/ttyUSB0)")
    parser.add_argument("benchdev", action="store", type=str,
                        help="serial port to act as test bench (e.g. /dev/ttyUSB1)")

    args = parser.parse_args()

    qmi.start("test_transport_serial", "")
    try:
        test_serial_transport(args.dutdev, args.benchdev, baudrate=115200)
    finally:
        qmi.stop()


if __name__ == "__main__":
    main()



from typing import NamedTuple, List, Callable, Tuple

import numpy as np
from numpy.typing import ArrayLike

from qmi.instruments.picoquant import SYNC_TYPE

NUM_CHANNELS = 8


class RealTimeHistogram(NamedTuple):
    """Real-time histogram from MultiHarp T2 event data.

    Attributes:
        start_timestamp: MultiHarp timestamp of first SYNC event of the histogram.
        bin_resolution:  Histogram bin resolution as multiple of the instrument base resolution.
        num_sync:        Number of SYNC periods included in this histogram.
        channels:        List of channel numbers included in the histogram.
        histogram_data:  2D array of shape (num_channels, num_bins) containing counts for each bin.
    """
    start_timestamp: int
    bin_resolution: int
    num_sync: int
    channels: List[int]
    histogram_data: np.ndarray


class RealTimeCountRate(NamedTuple):
    """Real-time count rate from MultiHarp T2 event data.

    Attributes:
        start_timestamp: MultiHarp timestamp of the SYNC event that started the counters.
        num_sync:        Number of SYNC periods included in the counter values.
        counts:          1D array of counter value for each input channel.
    """
    start_timestamp: int
    num_sync: int
    counts: np.ndarray


class _RealTimeHistogramProcessor:
    """Extract real-time histograms and count rates from event data."""

    def __init__(self, publish_histogram_func: Callable, publish_countrate_func: Callable) -> None:
        """Initialize histogram processing.

        Parameters:
            publish_histogram_func: A call-back function for publishing the histogram.
            publish_countrate_func: A call-back function for publishing the count rate.
        """
        self._publish_histogram_func = publish_histogram_func
        self._publish_countrate_func = publish_countrate_func
        self._previous_sync_timestamp = -1

        self._histogram_channels: List[int] = []
        self._histogram_resolution = 1
        self._histogram_num_bins = 0
        self._histogram_num_sync = 0
        self._histogram_data = np.zeros((0, 0), dtype=np.uint32)
        self._histogram_sync_counter = -1
        self._histogram_start_timestamp = 0

        self._countrate_aperture = (0, 0)
        self._countrate_num_sync = 0
        self._countrate_data = np.zeros(NUM_CHANNELS, dtype=np.uint64)
        self._countrate_sync_counter = -1
        self._countrate_start_timestamp = 0

    def set_histogram_config(self, channels: List[int], bin_resolution_ps: int, num_bins: int, num_sync: int) -> None:
        """Configure real-time histograms.

        Parameters:
            channels: Channel numbers to collect the histogram for.
            bin_resolution_ps: The bin resolution in picoseconds.
            num_sync: The number of synchronizations to collect the histogram data for.

        """

        assert bin_resolution_ps > 0
        self._histogram_channels = channels
        self._histogram_resolution = bin_resolution_ps
        self._histogram_num_bins = num_bins
        self._histogram_num_sync = num_sync

        # Clear data.
        self._histogram_data = np.zeros((len(channels), num_bins), dtype=np.uint32)
        self._histogram_sync_counter = -1

    def set_countrate_config(self, sync_aperture: Tuple[int, int], num_sync: int) -> None:
        """Configure real-time count rate reporting.

        Parameters:
            sync_aperture: [min, max] aperture times respective to the synchronization moment per channel.
            num_sync: The number of synchronizations to collect the count rate for.
        """
        self._countrate_aperture = sync_aperture
        self._countrate_num_sync = num_sync

        # Clear data.
        self._countrate_data = np.zeros(NUM_CHANNELS, dtype=np.uint64)
        self._countrate_sync_counter = -1

    def process_events(self, event_records: np.ndarray) -> None:
        """Process event records and update histogram and count rates.

        Parameters:
            event_records: An array of (SYNC_TYPE, timestamp) event records.
        """

        # Do nothing if histogram and count rate features are both disabled.
        if (self._histogram_num_sync == 0) and (self._countrate_num_sync == 0):
            return

        # Prepend the previous SYNC event, if there is any.
        # This simplifies calculation of time deltas since last SYNC.
        previous_sync_inserted = False
        if self._previous_sync_timestamp >= 0:
            event_records = np.insert(event_records, 0, (SYNC_TYPE, self._previous_sync_timestamp))
            previous_sync_inserted = True

        # Find the indexes of SYNC events.
        (sync_events_idx,) = np.where(event_records["type"] == SYNC_TYPE)

        # Do nothing before we get the first SYNC event.
        if len(sync_events_idx) == 0:
            return

        # Update the timestamp of the last SYNC event.
        self._previous_sync_timestamp = event_records[sync_events_idx[-1]]["timestamp"]

        # For each event, determine time delta since the last SYNC event.
        # The result is only valid for event past the first SYNC.
        first_sync_idx = sync_events_idx[0]
        sync_deltas = _get_sync_deltas(event_records, sync_events_idx)

        # Discard all events before the first SYNC.
        # Also discard the previous SYNC, if we prepended it to the current dataset.
        if previous_sync_inserted:
            first_sync_idx += 1
            sync_events_idx = sync_events_idx[1:]
        event_records = event_records[first_sync_idx:]
        sync_deltas = sync_deltas[first_sync_idx:]
        sync_events_idx -= first_sync_idx

        # Update real-time histograms.
        if self._histogram_num_sync > 0:
            self._update_histogram(event_records, sync_events_idx, sync_deltas)

        # Update real-time count rates.
        if self._countrate_num_sync > 0:
            self._update_countrate(event_records, sync_events_idx, sync_deltas)

    def _update_histogram(self,
                          event_records: np.ndarray,
                          sync_events_idx: np.ndarray,
                          sync_deltas: np.ndarray
                          ) -> None:
        """Update real-time histograms."""

        channels = self._histogram_channels
        num_bins = self._histogram_num_bins
        num_sync = self._histogram_num_sync

        # Do nothing if histogram processing is disabled.
        if (num_sync < 1) or (num_bins < 1) or (len(channels) < 1):
            return

        # Convert delta time to histogram bin index.
        # Clip to maximum bin index.
        event_bins = sync_deltas // self._histogram_resolution
        event_bins = np.minimum(event_bins, num_bins - 1)

        # Skip events before the first SYNC.
        next_idx = 0
        if self._histogram_sync_counter < 0:
            if len(sync_events_idx) == 0:
                return
            next_idx = sync_events_idx[0]
            self._histogram_start_timestamp = event_records[next_idx]["timestamp"]

        # Visit each SYNC record which ends a histogram integration interval.
        event_types = event_records["type"]
        for sync_idx in sync_events_idx[num_sync-self._histogram_sync_counter-1::num_sync]:

            # Update histogram up to the end of the current integration interval.
            # Process each channel separately.
            for (chan_pos, chan) in enumerate(channels):
                chan_bool = (event_types[next_idx:sync_idx] == chan)
                chan_bins = event_bins[next_idx:sync_idx][chan_bool]
                # Note: input to bincount MUST be dtype int64
                #       output from bincount will be dtype int64
                self._histogram_data[chan_pos] += np.bincount(chan_bins.astype(np.int64),
                                                              minlength=num_bins
                                                              ).astype(np.uint32)

            # Publish completed histogram.
            self._publish_histogram_func(RealTimeHistogram(
                start_timestamp=self._histogram_start_timestamp,
                bin_resolution=self._histogram_resolution,
                num_sync=num_sync,
                channels=channels,
                histogram_data=self._histogram_data))

            # Reset histogram for new integration interval.
            self._histogram_data = np.zeros((len(channels), num_bins), dtype=np.uint32)
            self._histogram_start_timestamp = event_records[sync_idx]["timestamp"]
            next_idx = sync_idx

        # Update the SYNC counter.
        self._histogram_sync_counter = (self._histogram_sync_counter + len(sync_events_idx)) % num_sync

        # Update histogram for remaining events.
        # Process each channel separately.
        for (chan_pos, chan) in enumerate(channels):
            chan_bool = (event_types[next_idx:] == chan)
            chan_bins = event_bins[next_idx:][chan_bool]
            self._histogram_data[chan_pos] += np.bincount(chan_bins.astype(np.int64),
                                                          minlength=num_bins
                                                          ).astype(np.uint32)

    def _update_countrate(self,
                          event_records: np.ndarray,
                          sync_events_idx: np.ndarray,
                          sync_deltas: np.ndarray
                          ) -> None:
        """Update real-time count rates."""

        # Do nothing if count rate processing is disabled.
        num_sync = self._countrate_num_sync
        if num_sync < 1:
            return

        # Select events that pass the aperture filter.
        (delta_min, delta_max) = self._countrate_aperture
        events_selected_bool = ((sync_deltas >= delta_min) & (sync_deltas <= delta_max))

        # Narrow down to events on the normal input channels.
        event_types = event_records["type"]
        events_selected_bool &= (event_types < NUM_CHANNELS)

        # Skip events before the first SYNC.
        next_idx = 0
        if self._countrate_sync_counter < 0:
            if len(sync_events_idx) == 0:
                return
            next_idx = sync_events_idx[0]
            self._countrate_start_timestamp = event_records[next_idx]["timestamp"]

        # Visit each SYNC record which ends an integration interval.
        for sync_idx in sync_events_idx[num_sync-self._countrate_sync_counter-1::num_sync]:

            # Update counters up to the end of the current integration interval.
            event_chan = event_types[next_idx:sync_idx][events_selected_bool[next_idx:sync_idx]]
            # Note: input to bincount MUST be dtype int64
            #       output from bincount will be dtype int64
            self._countrate_data += np.bincount(event_chan.astype(np.int64),
                                                minlength=NUM_CHANNELS
                                                ).astype(np.uint64)

            # Publish integrated counter values.
            self._publish_countrate_func(RealTimeCountRate(
                start_timestamp=self._countrate_start_timestamp,
                num_sync=num_sync,
                counts=self._countrate_data))

            # Reset counters for new integration interval.
            self._countrate_data = np.zeros(NUM_CHANNELS, dtype=np.uint64)
            self._countrate_start_timestamp = event_records[sync_idx]["timestamp"]
            next_idx = sync_idx

        # Update the SYNC counter.
        self._countrate_sync_counter = (self._countrate_sync_counter + len(sync_events_idx)) % num_sync

        # Update counters for the remaining events.
        event_chan = event_types[next_idx:][events_selected_bool[next_idx:]]
        self._countrate_data += np.bincount(event_chan.astype(np.int64),
                                            minlength=NUM_CHANNELS
                                            ).astype(np.uint64)


def _get_sync_deltas(event_records: np.ndarray, sync_events_idx: np.ndarray) -> np.ndarray:
    """For each event record, determine the time delta since the last SYNC event.

    Parameters:
        event_records:   Array of structured event records with "type" and "timestamp" fields.
        sync_events_idx: Array of indexes into `event_records` corresponding to SYNC events.

    Returns:
        Array of time deltas since SYNC event for each event record.
        This array has the same length as `event_records`.
        The returned values for records occurring before the first SYNC record will be invalid.
    """

    # Extract timestamps of SYNC events.
    event_timestamps = event_records["timestamp"]
    sync_timestamps = event_timestamps[sync_events_idx]

    # If there is no SYNC record, return dummy invalid data.
    if len(sync_events_idx) == 0:
        return np.zeros_like(event_timestamps)

    # Find index of the first SYNC event.
    first_sync_idx = sync_events_idx[0]

    # Prepare an array holding, for each event record, the timestamp of the most recent SYNC event.
    most_recent_sync_timestamp = np.zeros_like(event_timestamps)
    most_recent_sync_timestamp[first_sync_idx] = sync_timestamps[0]
    most_recent_sync_timestamp[sync_events_idx[1:]] = np.diff(sync_timestamps)
    most_recent_sync_timestamp = np.cumsum(most_recent_sync_timestamp)

    # Prepare an array holding, for each event record, the elapsed time since the most recent SYNC event.
    sync_deltas = event_timestamps - most_recent_sync_timestamp

    return sync_deltas


class TttrHistogram:
    """Class to make histograms from Time Tagged Time Resolved (TTTR) T2-mode data."""

    def __init__(self, channel: int, numbins: int):
        """Initialization of histogram bins, counts and deltas. In T2 mode the default bin size is 1E-12 s.

        Parameters:
            channel: The channel number to collect TTTR histogram data for.
            numbins: The number of bins to divide the data for.
        """
        self.channel = channel
        self.bins = np.arange(numbins)
        self.bin_edges = np.arange(numbins + 1)

        self.counts = np.zeros(numbins, dtype=np.uint64)
        self.deltas = None
        self.previous = None  # No timestamp carried from previous events.

    def reset(self) -> None:
        """Reset counts, deltas and previous timestamp."""
        self.counts.fill(0)
        self.deltas = None
        self.previous = None  # No timestamp carried from previous events.

    def process(self, events: np.ndarray) -> None:
        """Process events, and filter by aperture (if applicable).

        Parameters:
            events: An array of (SYNC_TYPE, timestamp) event records.
        """

        sync_type = 64  # The type of SYNC events.

        # If we have a previously recorded SYNC timestamp, we prepend it to the EVENTS array.
        if self.previous is not None:
            events = np.insert(events, 0, (sync_type, self.previous))

        # Find the first SYNC event.
        first_sync_index, = np.where(events["type"] == sync_type)
        if len(first_sync_index) == 0:
            # No sync event found. Discard these events.
            return

        # Discard all events before the first SYNC event.
        first_sync_index = first_sync_index[0]
        if first_sync_index != 0:
            events = events[first_sync_index:]  # type: ignore

        # At this point, the first event is guaranteed to be a SYNC,
        #   which simplifies matters significantly.

        assert events["type"][0] == sync_type

        idx = np.where(events["type"] == sync_type)  # This includes the first event.
        timestamps = events["timestamp"][idx]  # The timestamps of the SYNC events.

        increments = np.insert(np.diff(timestamps), 0, events["timestamp"][0])

        most_recent_synctime = np.zeros_like(events["timestamp"])
        most_recent_synctime[idx] = increments
        most_recent_synctime = np.cumsum(most_recent_synctime)

        deltas = events["timestamp"] - most_recent_synctime

        deltas = deltas[events["type"] == self.channel]
        deltas = deltas[deltas < len(self.counts)]

        self.deltas = deltas

        hist, _ = np.histogram(deltas, self.bin_edges)

        hist = hist.astype(np.uint64)

        self.counts += hist

        self.previous = most_recent_synctime[-1]

    def get_plot_data(self, bin_resolution_s: float) -> Tuple[ArrayLike, ArrayLike, int]:
        """Convenience function for plotting to return bin values and counts to the caller.

        Parameters:
            bin_resolution_s: bin resolution for the histogram. Unit is seconds.

        Returns:
            Tuple of (bin values data, self.counts, number of bins in self.bins)
        """
        return self.bins * bin_resolution_s / 1e-9, self.counts, len(self.bins)



"""
Instrument driver for the Tektronix FCA3000 and compatible frequency counters.
"""

import logging
import math
import struct
from typing import List, Optional, Tuple

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method
from qmi.core.scpi_protocol import ScpiProtocol
from qmi.core.transport import create_transport

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class Tektronix_FCA3000(QMI_Instrument):
    """Instrument driver for the Tektronix FCA3000 frequency counter.

    This driver is also compatible with the Pendulum CNT-91.
    """

    # Default response timeout in seconds.
    DEFAULT_RESPONSE_TIMEOUT = 5.0

    def __init__(self, context: QMI_Context, name: str, transport: str) -> None:
        """Initialize the instrument driver.

        :param name: Name for this instrument instance.
        :param transport: Transport descriptor to access this instrument.
            For the Pendulum CNT-91, this is typically
            "usbtmc:vendorid=0x14eb:productid=0x0x0091:serialnr=<device_serial_number>"
        """

        super().__init__(context, name)
        self._transport = create_transport(transport)
        self._scpi_transport = ScpiProtocol(self._transport)

    @rpc_method
    def open(self) -> None:
        _logger.info("Opening connection to %s", self._name)
        self._transport.open()
        super().open()

    @rpc_method
    def close(self) -> None:
        _logger.info("Closing connection to %s", self._name)
        super().close()
        self._transport.close()

    @rpc_method
    def reset(self) -> None:
        """Reset the instrument, returning (most) settings to their defaults."""
        self._scpi_transport.write("*RST")
        self._scpi_transport.ask("*OPC?")

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """Read instrument type and version and return QMI_InstrumentIdentification instance."""
        resp = self._scpi_transport.ask("*IDN?")
        words = resp.rstrip().split(",")
        if len(words) != 4:
            raise QMI_InstrumentException(f"Unexpected response to *IDN?, got {resp!r}")
        return QMI_InstrumentIdentification(
            vendor=words[0].strip(),
            model=words[1].strip(),
            serial=words[2].strip(),
            version=words[3].strip(),
        )

    @rpc_method
    def get_errors(self) -> List[str]:
        """Read pending error messages from the instrument."""
        errors = []
        while True:
            resp = self._scpi_transport.ask("SYST:ERR?")
            if resp.startswith("0,"):
                break
            errors.append(resp)
        return errors

    @rpc_method
    def measure_frequency(self, channel: int = 1) -> float:
        """Start a single frequency measurement and return frequency [Hz].

        No configuration of the instrument is needed prior to this command.
        Each invocation of this command reconfigures the instrument.

        This command is easy to use but quite slow. It will typically take
        0.1 to 0.2 seconds to complete.

        Note: Attempting to read a frequency when there is no valid signal
        on the input of the frequency counter, may crash the USB connection
        to the instrument. When errors occur, the instrument must be
        power-cycled to recover.

        :param channel: Input channel to use (1 = channel A, 2 = channel B).
        """
        resp = self._scpi_transport.ask(f"MEAS:FREQ? (@{channel})")
        freq = float(resp.rstrip())
        return freq

    @rpc_method
    def configure_frequency(
        self,
        channel: int = 1,
        aperture: Optional[float] = None,
        trigger_level: Optional[float] = None,
        timestamp: bool = False,
    ) -> None:
        """Configure the instrument for frequency measurements.

        After this function returns, you can perform a series of measurements
        by calling read_value() or read_timestamped_value() in a loop.

        :param channel: Input channel for frequency measurements (1 = channel A, 2 = channel B).
        :param aperture: Acquisition time per measurement in seconds (default 0.01).
        :param trigger_level: Optional fixed trigger level.
            By default, the trigger level is set automatically at the beginning of each measurement.
            Setting a fixed trigger level makes the measurement faster.
        :param timestamp: True to enable timestamping of measurement values.
        """
        self._scpi_transport.write(f"CONF:FREQ (@{channel})")
        if aperture is not None:
            self._scpi_transport.write(f"ACQ:APER {aperture:.6g}")
        if trigger_level is not None:
            self._scpi_transport.write(f"INP{channel}:LEV:AUTO 0")
            self._scpi_transport.write(f"INP{channel}:LEV {trigger_level:.6g}")
        self._scpi_transport.write(f"FORM:TINF {int(timestamp)}")

    @rpc_method
    def read_value(self) -> float:
        """Read the next value in an ongoing measurement.

        Before calling this function, the instrument must be configured
        by calling configure_frequency().
        """
        response = self._scpi_transport.ask("READ?")
        words = response.split(",")
        return float(words[0])

    @rpc_method
    def read_timestamped_value(self) -> Tuple[float, float]:
        """Read a timestamped value in an ongoing measurement.

        Before calling this function, the instrument must be configured
        by calling configure_frequency() and timestamping must be enabled.

        :return: Tuple (timestamp, value) where "timestamp" is a relative timestamp in seconds.
        """
        response = self._scpi_transport.ask("READ?")
        words = response.split(",")
        if len(words) < 2:
            raise QMI_InstrumentException(
                f"Expecting timestamped value but got {response!r}"
            )
        val = float(words[0])
        timestamp = float(words[1])
        return timestamp, val

    @rpc_method
    def set_talk_only(self, enable: bool) -> None:
        """Enable or disable Talk Only mode.

        When the instrument is in talk only mode, the application must call
        get_value_talk_only() at short intervals to read each subsequent
        value from the instrument. No other commands to the instrument are
        allowed, except set_talk_only() to disable talk only mode.

        Usage example:

        .. code-block:: python

           instr.configure_frequency(channel=1, aperture=0.002, trigger_level=0)
           instr.set_talk_only(True)
           for i in range(n):
               (ts, freq) = instr.get_value_talk_only()
           instr.set_talk_only(False)

        This mode works only when the device is connected via GPIB, not USB.
        """
        self._check_is_open()

        # IFC - Interface Clear - to force the instrument out of Talk Only mode.
        # This is a special command which is interpreted by the Prologix
        # GPIB-ethernet adapter.
        self._scpi_transport.write("++ifc")

        # Read pending data in case the instrument was still sending.
        self._transport.read_until_timeout(nbytes=16384, timeout=1.0)

        # Check that the instrument is back to normal.
        self._scpi_transport.ask("*OPC?")

        if enable:
            # Enable binary mode (required for talk only).
            self._scpi_transport.write("FORM PACK")

            # Disable display (required for talk only).
            self._scpi_transport.write("DISP:ENAB 0")

            # Enable continuous initiating (required for talk only).
            self._scpi_transport.write("INIT:CONT 1")

            # Switch to talk only mode.
            self._scpi_transport.write("SYST:TALK 1")

            # Discard the first data value (it is usually invalid).
            self._scpi_transport.read_binary_data()

    @rpc_method
    def get_value_talk_only(self) -> Tuple[float, float]:
        """Read the next value in an ongoing measurement.

        Before calling this function, the instrument must be configured
        by calling configure_frequency() and then switched to talk only
        mode by calling set_talk_only(True).

        :return: Tuple (timestamp, value) where "timestamp" is a relative
            timestamp in seconds. If timestamping is not enabled in the
            instrument, a NaN value will be returned as timestamp.
        """
        resp = self._scpi_transport.read_binary_data()
        if len(resp) == 16:
            (val, timestamp) = struct.unpack(">dq", resp)
            timestamp = 1.0e-12 * timestamp
        else:
            (val,) = struct.unpack(">d", resp)
            timestamp = math.nan
        return (timestamp, val)

    @rpc_method
    def get_trigger_level(self, channel: int) -> float:
        """Read current trigger level used by the instrument.

        :param channel: Input channel (1 = channel A, 2 = channel B).
        :return: Trigger level in Volt.
        """
        response = self._scpi_transport.ask(f"INP{channel}:LEV?")
        return float(response)

    @rpc_method
    def set_display_enabled(self, enable: bool) -> None:
        """Enable or disable the display on the instrument operating panel.

        Measurements can be (slightly) faster when the display is disabled.

        :param enable: True to enable the display, False to disable it.
        """
        self._scpi_transport.write(f"DISP:ENAB {int(enable)}")

    @rpc_method
    def set_initiate_continuous(self, enable: bool) -> None:
        """Enable or disable continuous measurements.

        Enable continuous measurements to let the instrument spontaneously
        measure new values as fast as possible. This is useful to continuously
        refresh the value on the display of the operating panel while the
        instrument is not in use by software.

        Continuous measurements are automatically disabled when the
        instrument is configured to measure via software.

        :param enable: True to enable continous measurements.
        """
        self._scpi_transport.write(f"INIT:CONT {int(enable)}")



"""Unit-tests for Rohde&Schwarz base class."""
import logging
import unittest
from unittest.mock import call, patch

import qmi
from qmi.core.exceptions import QMI_InstrumentException, QMI_TimeoutException
from qmi.core.transport import QMI_TcpTransport
from qmi.instruments.rohde_schwarz.rs_base_signal_gen import RohdeSchwarz_Base


class TestRohdeSchwarzBase(unittest.TestCase):

    def setUp(self):
        logging.getLogger("qmi.instruments.rohde_schwarz.rs_base_signal_gen").setLevel(logging.CRITICAL)
        qmi.start("TestRohdeSchwarzBaseContext")
        # Add patches
        patcher = patch('qmi.instruments.rohde_schwarz.rs_base_signal_gen.create_transport', spec=QMI_TcpTransport)
        self._transport_mock = patcher.start().return_value
        self.addCleanup(patcher.stop)
        patcher2 = patch('qmi.instruments.rohde_schwarz.rs_base_signal_gen.ScpiProtocol', autospec=True)
        self._scpi_mock = patcher2.start().return_value
        self.addCleanup(patcher2.stop)
        # Make DUT
        self.instr: RohdeSchwarz_Base = RohdeSchwarz_Base(qmi.context(), "RSB100a", "")
        self.instr.open()

    def tearDown(self):
        self.instr.close()
        qmi.stop()
        logging.getLogger("qmi.instruments.rohde_schwarz.rs_base_signal_gen").setLevel(logging.NOTSET)

    def test_get_idn(self):
        """Test ident. """
        vendor = "vendor"
        model = "model"
        serial = "serial"
        version = "version"
        self._scpi_mock.ask.return_value = f"{vendor},{model},{serial},{version}"

        ident = self.instr.get_idn()

        self._scpi_mock.ask.assert_called_once_with("*IDN?")
        self.assertEqual(ident.vendor, vendor)
        self.assertEqual(ident.model, model)
        self.assertEqual(ident.serial, serial)
        self.assertEqual(ident.version, version)

    def test_wrong_idn_response(self):
        """Test ident. raises QMI_InstrumentException by a non-sense response."""
        self._scpi_mock.ask.return_value = "nonsense"

        with self.assertRaises(QMI_InstrumentException):
            self.instr.get_idn()

        self._scpi_mock.ask.assert_called_once_with("*IDN?")

    def test_get_idn_works_even_with_with_valueerror(self):
        """Test the case where the _internal_poll_calibration_finished gives ValueError."""
        resp = b"nonsense"
        self.instr._calibrating = True
        self._transport_mock.read_until.side_effect = [resp + b"\r\n"]
        self._scpi_mock.ask.return_value = "vendor,model,serial,version"

        self.instr.get_idn()

        self.assertEqual(self.instr._calibration_error, "Unexpected response to calibration command: {!r}".format(resp))

    def test_start_calibration_excepts(self):
        """Test start calibration excepts when previous calibration is still pending."""
        self._transport_mock.read_until.side_effect = [b"1\r\n"]
        self._scpi_mock.ask.return_value = "0,\"No error\""
        self.instr.start_calibration()
        with self.assertRaises(QMI_InstrumentException):
            self.instr.start_calibration()

    def test_check_calibrating_excepts_with_timeout(self):
        """Test the case where the internal _check_calibration call excepts."""
        self.instr._calibrating = True
        expected_exception = "Calibration in progress"
        self._transport_mock.read_until.side_effect = [QMI_TimeoutException]

        with self.assertRaises(QMI_InstrumentException) as exc:
            self.instr.get_idn()

        self.assertEqual(str(exc.exception), expected_exception)

    def test_reset(self):
        """Test reset."""
        self._scpi_mock.ask.return_value = "0,\"No error\""
        expected_calls = [
            call("*CLS"),
            call("*RST")
        ]

        self.instr.reset()

        self._scpi_mock.write.assert_has_calls(expected_calls)
        self._scpi_mock.ask.assert_called_once_with("SYST:ERR:ALL?")

    def test_error(self):
        """Test command with error response."""
        self._scpi_mock.ask.return_value = "1,\"Huge error\""
        expected_calls = [
            call("*CLS"),
            call("*RST")
        ]

        with self.assertRaises(QMI_InstrumentException):
            self.instr.reset()

        self._scpi_mock.write.assert_has_calls(expected_calls)
        self._scpi_mock.ask.assert_called_once_with("SYST:ERR:ALL?")

    def test_wrong_float_value(self):
        """Test wrong float value response."""
        self._scpi_mock.ask.return_value = "not a float"

        with self.assertRaises(QMI_InstrumentException):
            self.instr.get_frequency()

    def test_wrong_bool_value(self):
        """Test wrong bool value response."""
        self._scpi_mock.ask.return_value = "not a bool"

        with self.assertRaises(QMI_InstrumentException):
            self.instr.get_output_state()

    def test_poll_calibration_in_progress(self):
        """Test calibration state polling."""
        # Device doesn't respond during calibration
        self._transport_mock.read_until.side_effect = QMI_TimeoutException

        self.instr._calibrating = True
        result = self.instr.poll_calibration()

        self.assertIsNone(result)

    def test_poll_calibration_not_started(self):
        """Test calibration state polling."""
        with self.assertRaises(QMI_InstrumentException):
            self.instr.poll_calibration()

    def test_poll_calibration_result_ok(self):
        """Test calibration state polling."""
        # Calibration result.
        calibration_result = 123
        self._transport_mock.read_until.return_value = bytes(str(calibration_result).encode("ascii")) + b"\r\n"

        # Calibration state query response.
        self._scpi_mock.ask.return_value = "0,\"No error\""

        self.instr._calibrating = True
        result = self.instr.poll_calibration()

        self.assertEqual(result, calibration_result)
        self._scpi_mock.ask.assert_called_once_with("SYST:ERR:ALL?")

    def test_poll_calibration_error(self):
        """Test calibration state polling."""
        logging.getLogger("qmi.instruments.rohde_schwarz.rs_base_signal_gen").setLevel(logging.ERROR)
        # Calibration result.
        calibration_result = 456
        self._transport_mock.read_until.return_value = bytes(str(calibration_result).encode("ascii")) + b"\r\n"

        # Calibration state query response.
        self._scpi_mock.ask.return_value = "1,\"Some error\""

        self.instr._calibrating = True
        with self.assertRaises(QMI_InstrumentException):
            self.instr.poll_calibration()

        self._scpi_mock.ask.assert_called_once_with("SYST:ERR:ALL?")

    def test_get_set_frequency(self):
        """Test get/set frequency."""
        # Test get.
        value = 123.345
        self._scpi_mock.ask.return_value = value

        result = self.instr.get_frequency()

        self._scpi_mock.ask.assert_called_once_with(":FREQ?")
        self.assertEqual(result, value)

        # Test set.
        self._scpi_mock.ask.return_value = "0,\"No error\""
        target_value = 456.789

        self.instr.set_frequency(target_value)

        self._scpi_mock.write.assert_called_once_with(f":FREQ {target_value}")

    def test_get_set_phase(self):
        """Test get/set phase."""
        # Test get.
        value = 123.345
        self._scpi_mock.ask.return_value = value

        result = self.instr.get_phase()

        self._scpi_mock.ask.assert_called_once_with(":PHAS?")
        self.assertEqual(result, value)

        # Test set.
        self._scpi_mock.ask.return_value = "0,\"No error\""
        target_value = 456.789

        self.instr.set_phase(target_value)

        self._scpi_mock.write.assert_called_once_with(f":PHAS {target_value}")

    def test_get_set_power(self):
        """Test get/set power."""
        # Test get.
        value = 123.345
        self._scpi_mock.ask.return_value = value

        result = self.instr.get_power()

        self._scpi_mock.ask.assert_called_once_with(":POW?")
        self.assertEqual(result, value)

        # Test set.
        self._scpi_mock.ask.return_value = "0,\"No error\""
        target_value = 456.789

        self.instr.set_power(target_value)

        self._scpi_mock.write.assert_called_once_with(f":POW {target_value}")

    def test_get_set_output_state(self):
        """Test get/set output state."""
        # Test get.
        value = "0"
        self._scpi_mock.ask.return_value = value

        result = self.instr.get_output_state()

        self._scpi_mock.ask.assert_called_once_with(":OUTP?")
        self.assertEqual(result, False)

        # Test set.
        self._scpi_mock.ask.return_value = "0,\"No error\""
        for target_value in (True, False):
            self.instr.set_output_state(target_value)

            self._scpi_mock.write.assert_called_once_with(":OUTP {}".format(
                1 if target_value else 0
            ))
            self._scpi_mock.write.reset_mock()

    def test_get_set_output_policy(self):
        """Test get/set output policy."""
        # Test get.
        value = "OFF"
        self._scpi_mock.ask.return_value = value

        result = self.instr.get_power_on_output_policy()

        self._scpi_mock.ask.assert_called_once_with(":OUTP:PON?")
        self.assertEqual(result, value)

        # Test set.
        self._scpi_mock.ask.return_value = "0,\"No error\""
        for target_value in ("OFF", "off", "UNCH", "unch"):
            self.instr.set_power_on_output_policy(target_value)

            self._scpi_mock.write.assert_called_once_with(f":OUTP:PON {target_value.upper()}")
            self._scpi_mock.write.reset_mock()

        # Test invalid value.
        with self.assertRaises(ValueError):
            self.instr.set_power_on_output_policy("chaos")

    def test_get_set_ref_source(self):
        """Test get/set reference source."""
        # Test get.
        value = "INT"
        self._scpi_mock.ask.return_value = value

        result = self.instr.get_reference_source()

        self._scpi_mock.ask.assert_called_once_with(":ROSC:SOUR?")
        self.assertEqual(result, value)

        # Test set.
        self._scpi_mock.ask.return_value = "0,\"No error\""
        for target_value in ("INT", "int", "EXT", "ext"):
            self.instr.set_reference_source(target_value)

            self._scpi_mock.write.assert_called_once_with(f":ROSC:SOUR {target_value.upper()}")
            self._scpi_mock.write.reset_mock()

        # Test invalid value.
        with self.assertRaises(ValueError):
            self.instr.set_reference_source("theoracle")

    def test_get_set_pulsemod_enable(self):
        """Test get/set pulse modulation."""
        # Test get.
        value = "0"
        self._scpi_mock.ask.return_value = value

        result = self.instr.get_pulsemod_enabled()

        self._scpi_mock.ask.assert_called_once_with(":PULM:STAT?")
        self.assertEqual(result, False)

        # Test set.
        self._scpi_mock.ask.return_value = "0,\"No error\""
        for target_value in (True, False):
            self.instr.set_pulsemod_enabled(target_value)

            self._scpi_mock.write.assert_called_once_with(":PULM:STAT {}".format(
                1 if target_value else 0
            ))
            self._scpi_mock.write.reset_mock()

    def test_get_set_pulsemod_ext_source(self):
        """Test get/set external pulse modulation source."""
        # Test get.
        value = "INT"
        self._scpi_mock.ask.return_value = value

        result = self.instr.get_pulsemod_ext_source()

        self._scpi_mock.ask.assert_called_once_with(":PULM:SOUR?")
        self.assertEqual(result, False)

        # Test set.
        self._scpi_mock.ask.return_value = "0,\"No error\""
        for target_value in (True, False):
            self.instr.set_pulsemod_ext_source(target_value)

            self._scpi_mock.write.assert_called_once_with(":PULM:SOUR {}".format(
                "EXT" if target_value else "INT"
            ))
            self._scpi_mock.write.reset_mock()

    def test_get_set_pulsemod_ext_source_excepts(self):
        """Test get external pulse modulation source excepts."""
        value = "ENT"
        self._scpi_mock.ask.return_value = value

        with self.assertRaises(QMI_InstrumentException):
            self.instr.get_pulsemod_ext_source()

    def test_get_set_pulsemod_polarity(self):
        """Test get/set pulse modulation polarity."""
        # Test get.
        value = "INV"
        self._scpi_mock.ask.return_value = value

        result = self.instr.get_pulsemod_polarity()

        self._scpi_mock.ask.assert_called_once_with(":PULM:POL?")
        self.assertEqual(result, True)

        # Test set.
        self._scpi_mock.ask.return_value = "0,\"No error\""
        for target_value in (True, False):
            self.instr.set_pulsemod_polarity(target_value)

            self._scpi_mock.write.assert_called_once_with(":PULM:POL {}".format(
                "INV" if target_value else "NORM"
            ))
            self._scpi_mock.write.reset_mock()

    def test_get_pulsemod_polarity_excepts(self):
        """Test get pulse modulation polarity excepts."""
        # Test get.
        value = "NINV"
        self._scpi_mock.ask.return_value = value

        with self.assertRaises(QMI_InstrumentException):
            self.instr.get_pulsemod_polarity()

    def test_get_set_iq_enable(self):
        """Test get/set IQ enable."""
        # Test get.
        value = "1"
        self._scpi_mock.ask.return_value = value

        result = self.instr.get_iq_enabled()

        self._scpi_mock.ask.assert_called_once_with(":IQ:STAT?")
        self.assertEqual(result, True)

        # Test set.
        self._scpi_mock.ask.return_value = "0,\"No error\""
        for target_value in (True, False):
            self.instr.set_iq_enabled(target_value)

            self._scpi_mock.write.assert_called_once_with(":IQ:STAT {}".format(
                1 if target_value else 0
            ))
            self._scpi_mock.write.reset_mock()

    def test_get_set_iq_wideband(self):
        """Test get/set IQ wideband."""
        # Test get.
        value = "0"
        self._scpi_mock.ask.return_value = value

        result = self.instr.get_iq_wideband()

        self._scpi_mock.ask.assert_called_once_with(":IQ:WBST?")
        self.assertEqual(result, False)

        # Test set.
        self._scpi_mock.ask.return_value = "0,\"No error\""
        for target_value in (True, False):
            self.instr.set_iq_wideband(target_value)

            self._scpi_mock.write.assert_called_once_with(":IQ:WBST {}".format(
                1 if target_value else 0
            ))
            self._scpi_mock.write.reset_mock()

    def test_get_set_iq_crest(self):
        """Test get/set IQ crest factor."""
        # Test get.
        value = 123.345
        self._scpi_mock.ask.return_value = value

        result = self.instr.get_iq_crest_factor()

        self._scpi_mock.ask.assert_called_once_with(":IQ:CRES?")
        self.assertEqual(result, value)

        # Test set.
        self._scpi_mock.ask.return_value = "0,\"No error\""
        target_value = 456.789

        self.instr.set_iq_crest_factor(target_value)

        self._scpi_mock.write.assert_called_once_with(f":IQ:CRES {target_value}")

    def test_get_set_iq_quad_offset(self):
        """Test get/set IQ quadrature offset."""
        # Test get.
        value = 1.2345
        self._scpi_mock.ask.return_value = value

        result = self.instr.get_iq_quadrature_offset()

        self._scpi_mock.ask.assert_called_once_with(":IQ:IMP:QUAD?")
        self.assertEqual(result, value)

        # Test set.
        self._scpi_mock.ask.return_value = "0,\"No error\""
        for target_value in [-8, -2, 0, 2, 8]:
            self.instr.set_iq_quadrature_offset(target_value)

            self._scpi_mock.write.assert_called_once_with(f":IQ:IMP:QUAD {target_value:.2f}")
            self._scpi_mock.write.reset_mock()

        # Test invalid value.
        with self.assertRaises(ValueError):
            self.instr.set_iq_quadrature_offset(-10)

        with self.assertRaises(ValueError):
            self.instr.set_iq_quadrature_offset(10)

    def test_get_set_iq_leakage_i(self):
        """Test get/set IQ leakage."""
        # Test get.
        value = 1.2345
        self._scpi_mock.ask.return_value = value

        result = self.instr.get_iq_leakage_i()

        self._scpi_mock.ask.assert_called_once_with(":IQ:IMP:LEAK:I?")
        self.assertEqual(result, value)

        # Test set.
        self._scpi_mock.ask.return_value = "0,\"No error\""
        for target_value in [-5, -2, 0, 2, 5]:
            self.instr.set_iq_leakage_i(target_value)

            self._scpi_mock.write.assert_called_once_with(f":IQ:IMP:LEAK:I {target_value:.2f}")
            self._scpi_mock.write.reset_mock()

        # Test invalid value.
        with self.assertRaises(ValueError):
            self.instr.set_iq_leakage_i(-6)

        with self.assertRaises(ValueError):
            self.instr.set_iq_leakage_i(6)

    def test_get_set_iq_leakage_q(self):
        """Test get/set IQ leakage."""
        # Test get.
        value = 1.2345
        self._scpi_mock.ask.return_value = value

        result = self.instr.get_iq_leakage_q()

        self._scpi_mock.ask.assert_called_once_with(":IQ:IMP:LEAK:Q?")
        self.assertEqual(result, value)

        # Test set.
        self._scpi_mock.ask.return_value = "0,\"No error\""
        for target_value in [-5, -2, 0, 2, 5]:
            self.instr.set_iq_leakage_q(target_value)

            self._scpi_mock.write.assert_called_once_with(f":IQ:IMP:LEAK:Q {target_value:.2f}")
            self._scpi_mock.write.reset_mock()

        # Test invalid value.
        with self.assertRaises(ValueError):
            self.instr.set_iq_leakage_q(-6)

        with self.assertRaises(ValueError):
            self.instr.set_iq_leakage_q(6)

    def test_get_set_iq_imbalance(self):
        """Test get/set IQ imbalance."""
        # Test get.
        value = 1.2345
        self._scpi_mock.ask.return_value = value

        result = self.instr.get_iq_gain_imbalance()

        self._scpi_mock.ask.assert_called_once_with(":IQ:IMP:IQR:MAGN?")
        self.assertEqual(result, value)

        # Test set.
        self._scpi_mock.ask.return_value = "0,\"No error\""
        for target_value in [-1, -0.5, 0, 0.5, 1]:
            self.instr.set_iq_gain_imbalance(target_value)

            self._scpi_mock.write.assert_called_once_with(f":IQ:IMP:IQR:MAGN {target_value:.3f}")
            self._scpi_mock.write.reset_mock()

        # Test invalid value.
        with self.assertRaises(ValueError):
            self.instr.set_iq_gain_imbalance(-2)

        with self.assertRaises(ValueError):
            self.instr.set_iq_gain_imbalance(2)

    def test_get_errors_with_no_error(self):
        """Test no errors in queue."""
        self._scpi_mock.ask.return_value = "0,\"No error\""
        self.instr.get_errors()

        self._scpi_mock.ask.assert_called_once_with("SYST:ERR:ALL?")

    def test_get_errors_with_errors(self):
        """Test errors in queue."""
        self._scpi_mock.ask.return_value = "1,\"Some error\""
        with self.assertRaises(QMI_InstrumentException):
            self.instr.get_errors()

        self._scpi_mock.ask.assert_called_once_with("SYST:ERR:ALL?")

    def test_get_error_queue_length(self):
        """Test error queue length."""
        expected = 1
        self._scpi_mock.ask.return_value = expected
        actual = self.instr.get_error_queue_length()

        self.assertEqual(actual, expected)

        self._scpi_mock.ask.assert_called_once_with("SYST:ERR:COUN?")


class TestRohdeSchwarzBasePowerLimited(unittest.TestCase):

    def setUp(self):
        qmi.start("TestRohdeSchwarzBasePLContext")
        # Add patches
        patcher = patch('qmi.instruments.rohde_schwarz.rs_base_signal_gen.create_transport', spec=QMI_TcpTransport)
        self._transport_mock = patcher.start().return_value
        self.addCleanup(patcher.stop)
        patcher = patch('qmi.instruments.rohde_schwarz.rs_base_signal_gen.ScpiProtocol', autospec=True)
        self._scpi_mock = patcher.start().return_value
        self.addCleanup(patcher.stop)
        # Make DUT
        self.maxpower = 10  # max_continuous_power
        self.instr: RohdeSchwarz_Base = RohdeSchwarz_Base(qmi.context(), "RSB100a", "", self.maxpower)
        self.instr.open()

    def tearDown(self):
        self.instr.close()
        qmi.stop()

    def test_set_power_below_max(self):
        """Test set power with maximum power set."""
        self._scpi_mock.ask.side_effect = [
            "0,\"No error\""  # query to error state
        ]
        target_value = 0.5 * self.maxpower

        self.instr.set_power(target_value)

        self._scpi_mock.write.assert_called_once_with(f":POW {target_value}")

    def test_set_power_above_max_ok(self):
        """Test set power with maximum power set."""
        self._scpi_mock.ask.side_effect = [
            "ON",  # response to query pulsemod enabled
            "EXT",  # response to query pulsemod ext source
            "NORM",  # response to query pulsemod polarity
            "0,\"No error\""  # query to error state
        ]
        target_value = 2 * self.maxpower

        self.instr.set_power(target_value)

        self._scpi_mock.write.assert_called_once_with(f":POW {target_value}")

    def test_set_power_above_max_fail1(self):
        """Test set power with maximum power set."""
        self._scpi_mock.ask.side_effect = [
            "OFF",  # response to query pulsemod enabled
            "EXT",  # response to query pulsemod ext source
            "NORM",  # response to query pulsemod polarity
            "0,\"No error\""  # query to error state
        ]
        target_value = 2 * self.maxpower

        with self.assertRaises(QMI_InstrumentException):
            self.instr.set_power(target_value)

        self._scpi_mock.write.assert_not_called()

    def test_set_power_above_max_fail2(self):
        """Test set power with maximum power set."""
        self._scpi_mock.ask.side_effect = [
            "ON",  # response to query pulsemod enabled
            "INT",  # response to query pulsemod ext source
            "NORM",  # response to query pulsemod polarity
            "0,\"No error\""  # query to error state
        ]
        target_value = 2 * self.maxpower

        with self.assertRaises(QMI_InstrumentException):
            self.instr.set_power(target_value)

        self._scpi_mock.write.assert_not_called()

    def test_set_power_above_max_fail3(self):
        """Test set power with maximum power set."""
        self._scpi_mock.ask.side_effect = [
            "ON",  # response to query pulsemod enabled
            "EXT",  # response to query pulsemod ext source
            "INV",  # response to query pulsemod polarity
            "0,\"No error\""  # query to error state
        ]
        target_value = 2 * self.maxpower

        with self.assertRaises(QMI_InstrumentException):
            self.instr.set_power(target_value)

        self._scpi_mock.write.assert_not_called()

    def test_set_output_enable_below_max(self):
        """Test set output enable with maximum power set."""
        self._scpi_mock.ask.side_effect = [
            0.5 * self.maxpower,  # response to query power
            "0,\"No error\""  # query to error state
        ]

        self.instr.set_output_state(True)

        self._scpi_mock.write.assert_called_once_with(":OUTP 1")

    def test_set_output_enable_above_max_ok(self):
        """Test set output enable with maximum power set."""
        self._scpi_mock.ask.side_effect = [
            2 * self.maxpower,  # response to query power
            "ON",  # response to query pulsemod enabled
            "EXT",  # response to query pulsemod ext source
            "NORM",  # response to query pulsemod polarity
            "0,\"No error\""  # query to error state
        ]

        self.instr.set_output_state(True)

        self._scpi_mock.write.assert_called_once_with(":OUTP 1")

    def test_set_output_enable_above_max_fail1(self):
        """Test set output enable with maximum power set."""
        self._scpi_mock.ask.side_effect = [
            2 * self.maxpower,  # response to query power
            "OFF",  # response to query pulsemod enabled
            "EXT",  # response to query pulsemod ext source
            "NORM",  # response to query pulsemod polarity
            "0,\"No error\""  # query to error state
        ]

        with self.assertRaises(QMI_InstrumentException):
            self.instr.set_output_state(True)

        self._scpi_mock.write.assert_not_called()

    def test_set_output_enable_above_max_fail2(self):
        """Test set output enable with maximum power set."""
        self._scpi_mock.ask.side_effect = [
            2 * self.maxpower,  # response to query power
            "ON",  # response to query pulsemod enabled
            "INT",  # response to query pulsemod ext source
            "NORM",  # response to query pulsemod polarity
            "0,\"No error\""  # query to error state
        ]

        with self.assertRaises(QMI_InstrumentException):
            self.instr.set_output_state(True)

        self._scpi_mock.write.assert_not_called()

    def test_set_output_enable_above_max_fail3(self):
        """Test set output enable with maximum power set."""
        self._scpi_mock.ask.side_effect = [
            2 * self.maxpower,  # response to query power
            "ON",  # response to query pulsemod enabled
            "EXT",  # response to query pulsemod ext source
            "INV",  # response to query pulsemod polarity
            "0,\"No error\""  # query to error state
        ]

        with self.assertRaises(QMI_InstrumentException):
            self.instr.set_output_state(True)

        self._scpi_mock.write.assert_not_called()

    def test_set_pulsemod_enable_below_max(self):
        """Test set pulsemod enable with maximum power set."""
        self._scpi_mock.ask.side_effect = [
            0.5 * self.maxpower,  # response to query power
            "0,\"No error\""  # query to error state
        ]

        self.instr.set_pulsemod_enabled(False)

        self._scpi_mock.write.assert_called_once_with(":PULM:STAT 0")

    def test_set_pulsemod_enable_above_max_ok(self):
        """Test set pulsemod enable with maximum power set."""
        self._scpi_mock.ask.side_effect = [
            "0,\"No error\""  # query to error state
        ]

        self.instr.set_pulsemod_enabled(True)

        self._scpi_mock.write.assert_called_once_with(":PULM:STAT 1")

    def test_set_pulsemod_enable_above_max_fail(self):
        """Test set pulsemod enable with maximum power set."""
        self._scpi_mock.ask.side_effect = [
            2 * self.maxpower,  # response to query power
            "0,\"No error\""  # query to error state
        ]

        with self.assertRaises(QMI_InstrumentException):
            self.instr.set_pulsemod_enabled(False)

        self._scpi_mock.write.assert_not_called()

    def test_set_pulsemod_ext_source_below_max(self):
        """Test set pulsemod external source with maximum power set."""
        self._scpi_mock.ask.side_effect = [
            0.5 * self.maxpower,  # response to query power
            "0,\"No error\""  # query to error state
        ]

        self.instr.set_pulsemod_ext_source(False)

        self._scpi_mock.write.assert_called_once_with(":PULM:SOUR INT")

    def test_set_pulsemod_ext_source_above_max_ok(self):
        """Test set pulsemod external source with maximum power set."""
        self._scpi_mock.ask.side_effect = [
            "0,\"No error\""  # query to error state
        ]

        self.instr.set_pulsemod_ext_source(True)

        self._scpi_mock.write.assert_called_once_with(":PULM:SOUR EXT")

    def test_set_pulsemod_ext_source_above_max_fail(self):
        """Test set pulsemod enable with maximum power set."""
        self._scpi_mock.ask.side_effect = [
            2 * self.maxpower,  # response to query power
            "0,\"No error\""  # query to error state
        ]

        with self.assertRaises(QMI_InstrumentException):
            self.instr.set_pulsemod_ext_source(False)

        self._scpi_mock.write.assert_not_called()

    def test_set_pulsemod_polarity_below_max(self):
        """Test set pulsemod polarity with maximum power set."""
        self._scpi_mock.ask.side_effect = [
            0.5 * self.maxpower,  # response to query power
            "0,\"No error\""  # query to error state
        ]

        self.instr.set_pulsemod_polarity(True)

        self._scpi_mock.write.assert_called_once_with(":PULM:POL INV")

    def test_set_pulsemod_polarity_above_max_ok(self):
        """Test set pulsemod polarity with maximum power set."""
        self._scpi_mock.ask.side_effect = [
            "0,\"No error\""  # query to error state
        ]

        self.instr.set_pulsemod_polarity(False)

        self._scpi_mock.write.assert_called_once_with(":PULM:POL NORM")

    def test_set_pulsemod_polarity_above_max_fail(self):
        """Test set pulsemod enable with maximum power set."""
        self._scpi_mock.ask.side_effect = [
            2 * self.maxpower,  # response to query power
            "0,\"No error\""  # query to error state
        ]

        with self.assertRaises(QMI_InstrumentException):
            self.instr.set_pulsemod_polarity(True)

        self._scpi_mock.write.assert_not_called()


if __name__ == '__main__':
    unittest.main()



"""
Instrument driver for the Rigol DG4102 Function/Arbitrary Waveform Generator.
The instrument class is adapted from the very similar class Rohde und Schwarz SGS100A.
Written by Matthew Weaver and Yanik Herrmann (m.j.weaver@tudelft.nl).
"""

import logging
import re

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method
from qmi.core.scpi_protocol import ScpiProtocol
from qmi.core.transport import create_transport

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class Rigol_Dg4102(QMI_Instrument):
    """Instrument driver for the Rigol DG4102 Function/Arbitrary Waveform Generator."""

    # Default response timeout in seconds.
    DEFAULT_RESPONSE_TIMEOUT = 1.0

    def __init__(self,
                 context: QMI_Context,
                 name: str,
                 transport: str,
                 init_source: int = 1
                 ) -> None:
        """Initialize the instrument driver.

        Parameters:
            context:        The QMI context.
            name:           Name for this instrument instance.
            transport:      QMI transport descriptor to connect to the instrument.
            init_source:    The initial source to control. This can be changed via set_source.
        """
        super().__init__(context, name)
        self._timeout = self.DEFAULT_RESPONSE_TIMEOUT
        self._transport = create_transport(transport, default_attributes={"port": 5555})
        self._scpi_protocol = ScpiProtocol(self._transport, default_timeout=self._timeout)
        self._channel = init_source

    @rpc_method
    def open(self) -> None:
        _logger.info("[%s] Opening connection to instrument", self._name)
        super().open()
        self._transport.open()

    @rpc_method
    def close(self) -> None:
        _logger.info("[%s] Closing connection to instrument", self._name)
        self._transport.close()
        super().close()

    def _ask_float(self, cmd: str) -> float:
        """Send a query and return a floating point response.

        Parameters:
            cmd: The query string.
        """
        resp = self._scpi_protocol.ask(cmd)
        try:
            return float(resp)
        except ValueError:
            raise QMI_InstrumentException("Unexpected response to command {!r}: {!r}".format(cmd, resp))

    def _ask_bool(self, cmd: str) -> bool:
        """Send a query and return a boolean response.
        
        Parameters:
            cmd: The query string.
        """
        resp = self._scpi_protocol.ask(cmd)
        value = resp.strip().upper()
        if value in ("1", "ON"):
            return True
        elif value in ("0", "OFF"):
            return False
        else:
            raise QMI_InstrumentException("Unexpected response to command {!r}: {!r}".format(cmd, resp))

    def _check_error(self) -> None:
        """Read the instrument error queue and raise an exception if there is an error.
        
        Raises:
            QMI_InstrumentException if response error code is not 0.
        """
        resp = self._scpi_protocol.ask("SYSTem:ERRor?")
        # When there are no errors, the response is '0,"No error"'.
        if not re.match(r"^\s*0\s*,", resp):
            # Some error occurred.
            raise QMI_InstrumentException("Instrument returned error: {}".format(resp))

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """Read instrument type and version. 
        
        Returns:
            QMI_InstrumentIdentification instance.
        """
        resp = self._scpi_protocol.ask("*IDN?")
        words = resp.rstrip().split(",")
        if len(words) != 4:
            raise QMI_InstrumentException("Unexpected response to *IDN?, got {!r}".format(resp))
        return QMI_InstrumentIdentification(vendor=words[0].strip(),
                                            model=words[1].strip(),
                                            serial=words[2].strip(),
                                            version=words[3].strip())

    @rpc_method
    def set_source(self, new_source: int) -> None:
        """Choose which source channel to control.

        Parameters:
            new_source: The new source channel number.
        """
        self._channel = new_source

    @rpc_method
    def get_source(self) -> int:
        """Return the source channel which is currently controlled

        Returns:
            The current source channel number.
        """
        return self._channel

    @rpc_method
    def get_output_state(self) -> bool:
        """Return True if output is enabled, False if output is disabled.

        Returns:
            output_state: The output state.
        """
        output_state = self._ask_bool(":OUTP{}?".format(self._channel))
        # For some reason the response to OUTP? terminates with '\n\n' instead of the usual '\n'.
        # This clears the extra empty line.
        self._scpi_protocol.ask("")
        return output_state

    @rpc_method
    def set_output_state(self, enable: bool) -> None:
        """Enable or disable output

        Parameters:
            enable: The new output state. True for "ON", False for "OFF".
        """
        self._scpi_protocol.write(":OUTP{} {}".format(self._channel, "ON" if enable else "OFF"))
        self._check_error()

    @rpc_method
    def get_waveform(self) -> str:
        """Return the current waveform.

        Returns:
            The current waveform text.
        """
        cmd = ":SOURce{}:FUNCtion?".format(self._channel)
        return self._scpi_protocol.ask(cmd)

    @rpc_method
    def set_waveform(self, waveform: str) -> None:
        """Set the waveform of a channel.

        Parameters:
            waveform: The new waveform type. Allowed waveforms are:
                      SINusoid|SQUare|RAMP|PULSe|NOISe|USER|HARMonic|CUSTom|DC
        """
        self._scpi_protocol.write(":SOURce{}:FUNCtion {}".format(self._channel, waveform))
        self._check_error()

    @rpc_method
    def get_frequency(self) -> float:
        """Return the current frequency of a channel in Hz.

        Returns:
            The frequency of a channel.
        """
        return self._ask_float(":SOURce{}:FREQ?".format(self._channel))

    @rpc_method
    def set_frequency(self, frequency: float) -> None:
        """Set the frequency of a channel in Hz.

        Parameters:
            frequency: The new channel frequency.
        """
        self._scpi_protocol.write(":SOURce{}:FREQ {}".format(self._channel, frequency))
        self._check_error()

    @rpc_method
    def get_amplitude(self) -> float:
        """Return the current amplitude of a channel in V.

        Returns:
            The channel amplitude.
        """
        return self._ask_float(":SOURce{}:VOLT?".format(self._channel))

    @rpc_method
    def set_amplitude(self, amplitude: float) -> None:
        """Set the amplitude of a channel in V.

        Parameters:
            amplitude: The new channel amplitude.
        """
        self._scpi_protocol.write(":SOURce{}:VOLT {}".format(self._channel, amplitude))
        self._check_error()

    @rpc_method
    def get_offset(self) -> float:
        """Return the current offset of a channel in V.

        Returns:
            The channel offset voltage.
        """
        return self._ask_float(":SOURce{}:VOLTage:OFFSet?".format(self._channel))

    @rpc_method
    def set_offset(self, offset: float) -> None:
        """Set the offset of a channel in V.

        Parameters:
            offset: The new channel offset voltage.
        """
        self._scpi_protocol.write(":SOURce{}:VOLTage:OFFSet {}".format(self._channel, offset))
        self._check_error()

    @rpc_method
    def get_phase(self) -> float:
        """Return the current phase of a channel in deg.

        Returns:
            The channel phase.
        """
        return self._ask_float(":SOURce{}:PHASe?".format(self._channel))

    @rpc_method
    def set_phase(self, phase: float) -> None:
        """Set the phase of a channel in deg.

        Parameters:
            phase: The new channel phase.
        """
        self._scpi_protocol.write(":SOURce{}:PHASe {}".format(self._channel, phase))
        self._check_error()



from unittest import TestCase
from unittest.mock import patch, sentinel

from qmi.instruments.high_finesse.support import wlmData
from qmi.instruments.high_finesse.support._library_wrapper import _LibWrapper


class TestHighFinesse_WS_Linux(TestCase):
    """Test the behavior of the High Finesse driver on a Linux platform."""

    def setUp(self):
        """Make the sys mock say that we run on Linux."""
        patcher = patch("qmi.instruments.high_finesse.support._library_wrapper.sys")
        self.sys_mock = patcher.start()
        self.sys_mock.configure_mock(platform="linux")
        self.addCleanup(patcher.stop)

    @patch("qmi.instruments.high_finesse.support._library_wrapper.wlmData.LoadDLL")
    def test_init(self, mocker):
        """The LoadDLL function should be called once."""
        wlmData.dll = sentinel.dll

        lib = _LibWrapper()
        mocker.assert_called_once_with("libwlmData.so")
        self.assertIsNotNone(lib.dll)
        self.assertEqual(lib.dll, sentinel.dll)


class TestHighFinesse_WS_Windows(TestCase):
    """Test the behavior of the High Finesse driver on a Windows platform."""

    def setUp(self):
        """Make the sys mock say that we run on Windows."""
        patcher = patch("qmi.instruments.high_finesse.support._library_wrapper.sys")
        self.sys_mock = patcher.start()
        self.sys_mock.configure_mock(platform="windows")
        self.addCleanup(patcher.stop)

    @patch("qmi.instruments.high_finesse.support._library_wrapper.wlmData.LoadDLL")
    def test_init(self, mocker):
        """The LoadDLL function should be called once."""
        wlmData.dll = sentinel.dll

        lib = _LibWrapper()
        mocker.assert_called_once_with("wlmData.dll")
        self.assertIsNotNone(lib.dll)
        self.assertEqual(lib.dll, sentinel.dll)


class TestHighFinesse_WS_MacOS(TestCase):
    """Test the behavior of the High Finesse driver on a MacOS platform."""

    def setUp(self):
        """Make the sys mock say that we run on MacOS."""
        patcher = patch("qmi.instruments.high_finesse.support._library_wrapper.sys")
        self.sys_mock = patcher.start()
        self.sys_mock.configure_mock(platform="darwin")
        self.addCleanup(patcher.stop)

    @patch("qmi.instruments.high_finesse.support._library_wrapper.wlmData.LoadDLL")
    def test_init(self, mocker):
        """The LoadDLL function should be called once."""
        wlmData.dll = sentinel.dll

        lib = _LibWrapper()
        mocker.assert_called_once_with("libwlmData.dylib")
        self.assertIsNotNone(lib.dll)
        self.assertEqual(lib.dll, sentinel.dll)


class TestHighFinesse_WS_Unknown(TestCase):
    """Test the behavior of the High Finesse driver on an unsupported platform."""

    def setUp(self):
        """Make the sys mock say that we run on Cygwin."""
        patcher = patch("qmi.instruments.high_finesse.support._library_wrapper.sys")
        self.sys_mock = patcher.start()
        self.sys_mock.configure_mock(platform="cygwin")
        self.addCleanup(patcher.stop)

    @patch("qmi.instruments.high_finesse.support._library_wrapper.wlmData.LoadDLL")
    def test_init(self, mocker):
        """The LoadDLL function should be called once."""
        with self.assertRaises(OSError) as exc:
            _ = _LibWrapper()

        self.assertEqual("Unsupported platform", str(exc.exception))
        mocker.assert_not_called()



"""
Zurich Instruments AG, arbitrary waveform generator.

The qmi.instruments.zurich_instruments package provides support for:
- HDAWG.
"""
from qmi.instruments.zurich_instruments.hdawg import CompilerStatus, UploadStatus
# Alternative, QMI naming convention approved name
from qmi.instruments.zurich_instruments.hdawg import ZurichInstruments_HDAWG as ZurichInstruments_Hdawg
